from abc import ABCMeta, abstractmethod, ABC
from re import L
from typing import Union, Optional, List

from ..rewriter import sql_rewriter
from ..rewriter.sql_generic import SqlNode

LITERAL_TYPE_STRING = "STRING"
LITERAL_TYPE_INTEGER = "INTEGER"
LITERAL_TYPE_DECIMAL = "DECIMAL"
LITERAL_TYPE_ARRAY = "ARRAY"


class Node(metaclass=ABCMeta):

    @abstractmethod
    def rewrite(self, rewriter: sql_rewriter.SqlRewriter) -> 'SqlNode': pass

    def __eq__(self, o: object) -> bool:
        return str(self).__eq__(str(o))

    def __hash__(self) -> int:
        return str(self).__hash__()


class Null(Node):

    def rewrite(self, rewriter: sql_rewriter.SqlRewriter) -> 'SqlNode':
        return rewriter.rewrite_null(self)


class Identifier(Node):

    def __init__(self, value: str, is_quoted: Optional[bool] = False):
        self._value = value
        self._is_quoted = is_quoted

    def value(self):
        if self._is_quoted:
            return self._value.strip('\"').strip('\'')
        return self._value

    def rewrite(self, rewriter: sql_rewriter.SqlRewriter) -> 'SqlNode':
        return rewriter.rewrite_identifier(self)

    def __str__(self):
        return self._value


class Literal(Node):

    def __init__(self, value: str, literal_type: str):
        self._value = value
        self.literal_type = literal_type

    def value(self):
        if self.literal_type == LITERAL_TYPE_STRING:
            return self._value.strip('\'')
        elif self.literal_type in [LITERAL_TYPE_INTEGER, LITERAL_TYPE_DECIMAL]:
            return str(self._value)

    def rewrite(self, rewriter: sql_rewriter.SqlRewriter) -> 'SqlNode':
        return rewriter.rewrite_literal(self)

    def __str__(self):
        return self._value


class StringLiteral(Literal):

    def __init__(self, str_value: str):
        super(StringLiteral, self).__init__(str_value, LITERAL_TYPE_STRING)


class IntegerLiteral(Literal):

    def __init__(self, int_value: Union[str, int]):
        super(IntegerLiteral, self).__init__(int_value, LITERAL_TYPE_INTEGER)


class DecimalLiteral(Literal):

    def __init__(self, decimal_value: Union[str, int]):
        super(DecimalLiteral, self).__init__(decimal_value, LITERAL_TYPE_DECIMAL)


class ArrayLiteral(Literal):

    def __init__(self, literals: List[Literal]):
        super(ArrayLiteral, self).__init__(",".join([l.value() for l in literals]), LITERAL_TYPE_ARRAY)
        self.literals = literals

    def value(self):
        return self.literals

    def rewrite(self, rewriter: sql_rewriter.SqlRewriter) -> 'SqlNode':
        return rewriter.rewrite_array_literal(self)


class QualifiedName(Node):

    def __init__(self, *parts: List[Union[Identifier, str]]):
        self.parts = [_auto_identifier(p) for p in parts]

    def rewrite(self, rewriter: sql_rewriter.SqlRewriter) -> 'SqlNode':
        raise NotImplemented

    def __str__(self):
        return '.'.join([p.value() for p in self.parts])


class View(Node):

    def __init__(self, name: Identifier):
        self.name = name

    def rewrite(self, rewriter: sql_rewriter.SqlRewriter) -> 'SqlNode':
        return rewriter.set_view(self)

    def __str__(self):
        return self.name


class Table(Node):

    def __init__(self, name: Identifier):
        self.name = name

    def rewrite(self, rewriter: sql_rewriter.SqlRewriter) -> 'SqlNode':
        return rewriter.rewrite_table(self)

    def __str__(self):
        return self.name


class Column(Node):

    def __init__(self, table: Table, name: Identifier, alias: Optional[Identifier] = None):
        self.table = table
        self.name = name
        self.alias = alias

    def rewrite(self, rewriter: sql_rewriter.SqlRewriter) -> 'SqlNode':
        return rewriter.rewrite_column(self)

    def __str__(self):
        return f"{self.table}.{self.name} AS {self.alias}"


class FunctionCall(Node):

    def __init__(self, name: str, args: List, alias: Optional[Identifier] = None):
        if args is not None and not isinstance(args, list):
            raise RuntimeError("args must be list or None")
        self.name = name
        self.args = args if args is not None else []
        self.alias = alias

    def rewrite(self, rewriter: sql_rewriter.SqlRewriter) -> 'SqlNode':
        return rewriter.rewrite_function_call(self)

    def __str__(self):
        return f"{self.name}({self.args}) AS {self.alias}"


class Predicate(Node, ABC):

    def __init__(self, operand: Union[Column, FunctionCall]):
        self.operand = operand


class Comparison(Predicate):

    def __init__(self, operand: Union[Column, FunctionCall], operator: str, literal: Literal):
        super(Comparison, self).__init__(operand)
        self.operator = operator
        self.literal = literal

    def rewrite(self, rewriter: sql_rewriter.SqlRewriter) -> 'SqlNode':
        return rewriter.rewrite_comparison(self)

    def __str__(self):
        return f"{self.operand} {self.operand} {self.literal}"


class IsNull(Predicate):

    def __init__(self, operand: Union[Column, FunctionCall], negated: bool):
        super(IsNull, self).__init__(operand)
        self.negated = negated

    def rewrite(self, rewriter: sql_rewriter.SqlRewriter) -> 'SqlNode':
        return rewriter.rewrite_is_null(self)

    def __str__(self):
        return f"{self.operand} IS {self.negated} NULL"


class Like(Predicate):

    def __init__(self, operand: Union[Column, FunctionCall], negated: bool, string_literal: Literal):
        super(Like, self).__init__(operand)
        self.negated = negated
        self.string_literal = string_literal

    def rewrite(self, rewriter: sql_rewriter.SqlRewriter) -> 'SqlNode':
        return rewriter.rewrite_like(self)

    def __str__(self):
        return f"{self.operand} {self.negated} LIKE {self.string_literal}"


class In(Predicate):

    def __init__(self, operand: Union[Column, FunctionCall], negated: bool, literals: List[Literal]):
        super(In, self).__init__(operand)
        self.negated = negated
        self.literals = literals

    def rewrite(self, rewriter: sql_rewriter.SqlRewriter) -> 'SqlNode':
        return rewriter.rewrite_in(self)

    def __str__(self):
        return f"{self.operand} {self.negated} IN ({self.literals})"


class Between(Predicate):

    def __init__(self, operand: Union[Column, FunctionCall], negated: bool, literal1: Literal, literal2: Literal):
        super(Between, self).__init__(operand)
        self.negated = negated
        self.literal1 = literal1
        self.literal2 = literal2

    def rewrite(self, rewriter: sql_rewriter.SqlRewriter) -> 'SqlNode':
        return rewriter.rewrite_between(self)

    def __str__(self):
        return f"{self.operand} {self.negated} BETWEEN {self.literal1} AND {self.literal2}"


class CaseExpression(Node):

    def __init__(self, whens_expressions: List[Node], else_expression: Node, alias: Optional[Identifier] = None):
        self.whens_expressions = whens_expressions
        self.else_expression = else_expression
        self.alias = alias

    def rewrite(self, rewriter: sql_rewriter.SqlRewriter) -> 'SqlNode':
        return rewriter.rewrite_case_expression(self)


class LogicalExpression(Node):

    def __init__(self, boolean_expressions: List[Predicate]):
        self.boolean_expressions = boolean_expressions

    def rewrite(self, rewriter: sql_rewriter.SqlRewriter) -> 'SqlNode':
        return rewriter.rewrite_logical_expression(self)

    def __getitem__(self, n):
        return self.boolean_expressions[n]

    def __len__(self):
        return len(self.boolean_expressions)

    def __iter__(self):
        return iter(self.boolean_expressions)


class WhenExpression(Node):

    def __init__(self, logical_expression: LogicalExpression, then_expression: Node):
        self.logical_expression = logical_expression
        self.then_expression = then_expression

    def rewrite(self, rewriter: sql_rewriter.SqlRewriter) -> 'SqlNode':
        return rewriter.rewrite_when_expression(self)


class OrderByItem(Node):

    def __init__(self, node: Node, direction: Optional[str] = None):
        self.node = node
        self.direction = 'ASC' if direction is None else direction

    def rewrite(self, rewriter: sql_rewriter.SqlRewriter) -> 'SqlNode':
        return rewriter.rewrite_order_by_item(self)

    def __str__(self):
        return f"{self.node} {self.direction}"


class Limit(Node):

    def __init__(self, literal: IntegerLiteral):
        self.literal = literal

    def rewrite(self, rewriter: sql_rewriter.SqlRewriter) -> 'SqlNode':
        return rewriter.rewrite_limit(self)

    def __str__(self):
        return f"{self.literal}"


class Offset(Node):

    def __init__(self, literal: IntegerLiteral):
        self.literal = literal

    def rewrite(self, rewriter: sql_rewriter.SqlRewriter) -> 'SqlNode':
        return rewriter.rewrite_offset(self)

    def __str__(self):
        return f"{self.literal}"


class BinaryExpression(Node):

    def __init__(self, lhs: Node, operator: str, rhs: Node, alias: Optional[Identifier] = None):
        self.lhs = lhs
        self.operator = operator
        self.rhs = rhs
        self.alias = alias

    def rewrite(self, rewriter: sql_rewriter.SqlRewriter) -> 'SqlNode':
        return rewriter.rewrite_binary_expression(self)

    def __str__(self):
        return f"{self.lhs} {self.operator} {self.rhs}"


class ParenExpression(Node):

    def __init__(self, binary_expr: BinaryExpression, alias: Optional[Identifier] = None):
        self.binary_expr = binary_expr
        self.alias = alias

    def rewrite(self, rewriter: sql_rewriter.SqlRewriter) -> 'SqlNode':
        return rewriter.rewrite_paren_expression(self)

    def __str__(self):
        return f"({self.binary_expr})"


class FlatQL:

    def __init__(self, projections: List[Union[QualifiedName, FunctionCall]], view: View,
                 predicates: Optional[LogicalExpression] = None, order_by_items: Optional[List[Node]] = None,
                 limit: Optional[Limit] = None, offset: Optional[Offset] = None):
        self.projections = projections
        self.view = view
        self.predicates = predicates
        self.order_by_items = order_by_items
        self.limit = limit
        self.offset = offset

    def rewrite(self, rewriter: sql_rewriter.SqlRewriter):
        rewriter.set_view(self.view)

        if self.predicates is not None:
            rewriter.set_predicates(self.predicates.rewrite(rewriter))

        if self.order_by_items is not None:
            rewriter.set_order_by_items([o.rewrite(rewriter) for o in self.order_by_items])

        if self.limit is not None:
            rewriter.set_limit(self.limit.rewrite(rewriter))

        if self.offset is not None:
            rewriter.set_offset(self.offset.rewrite(rewriter))

        rewriter.set_projections([p.rewrite(rewriter) for p in self.projections])


def _auto_identifier(raw_identifier: Union[Identifier, str]):
    if isinstance(raw_identifier, str):
        return Identifier(raw_identifier)
    if isinstance(raw_identifier, Identifier):
        return raw_identifier
    else:
        return Identifier(str(raw_identifier))
