import re
import types
from typing import List, Union

import ply.lex as lex
import ply.yacc as yacc

from .ast import FlatQL, QualifiedName, Literal, FunctionCall, Comparison, IsNull, Like, In, Between, \
    Identifier, OrderByItem, View, Column, Table, Limit, Offset, IntegerLiteral, Predicate, BinaryExpression, \
    ParenExpression, WhenExpression, CaseExpression, Null, LogicalExpression, ArrayLiteral

QUALIFIED_NAME_PATTERN = r"[a-zA-Z_][a-zA-Z0-9_@:]*(\.[a-zA-Z_][a-zA-Z0-9_@:]*)*"


# noinspection PySingleQuotedDocstring
class FlatQLParser:
    tokens = (
        'SELECT', 'FROM', 'WHERE', 'AND', 'BETWEEN', 'IN', 'LIKE', 'NOT', 'IS', 'NULL', 'AS',
        'ORDER', 'BY', 'OFFSET', 'LIMIT', 'DESC', 'ASC',
        'STRING', 'NUMBER', 'IDENTIFIER', 'QUOTED_IDENTIFIER', 'INTEGER', 'DECIMAL',
        'LBRACK', 'RBRACK', 'LPAREN', 'RPAREN', 'GT', 'GE', 'LT', 'LE', 'EQ', 'NE', 'PLUS', 'MINUS', 'EXP', 'TIMES', 'DIVIDE',
        'COMMA', 'PERIOD', 'CASE', 'WHEN', 'THEN', 'ELSE', 'END'
    )
    reserved = sorted(set(tokens))

    t_LPAREN = r'\('
    t_RPAREN = r'\)'
    t_LBRACK = r'\['
    t_RBRACK = r'\]'

    t_EQ = r'='
    t_NE = r'!='
    t_LT = r'<'
    t_LE = r'<='
    t_GT = r'>'
    t_GE = r'>='

    t_PLUS = r'\+'
    t_MINUS = r'-'
    t_EXP = r'\*\*'
    t_TIMES = r'\*'
    t_DIVIDE = r'/'

    t_PERIOD = r'\.'
    t_COMMA = r'\,'
    t_STRING = r"'([^']|'')*'"

    t_ignore = ' \t'

    def __init__(self):
        lex.lex(module=self, reflags=re.IGNORECASE)
        yacc.yacc(module=self, write_tables=False, debug=False)

    def t_NUMBER(self, t):
        r"""[-+]?(\d+(?:\.\d*)?(?:[eE][+-]?\d+)?|\d*(?:\.\d+)(?:[eE][+-]?\d+)?)"""
        if 'e' in t.value or 'E' in t.value or '.' in t.value:
            t.type = 'DECIMAL'
        else:
            t.type = 'INTEGER'
        return t

    def t_INTEGER(self, t):
        r'\d+'
        t.type = "INTEGER"
        return t

    def t_comment(self, t):
        r'(/\*(.|\n)*?\*/)|(--[^\r\n]*\r?\n?)'
        t.lexer.lineno += t.value.count('\n')

    def t_IDENTIFIER(self, t):
        r"""[a-zA-Z_][a-zA-Z0-9_]*"""
        val = t.value.lower()
        if val.upper() in self.__class__.reserved:
            t.type = val.upper()
        return t

    def t_QUOTED_IDENTIFIER(self, t):
        r'"([^"]|"")*"'
        val = t.value.lower()
        if val in self.__class__.reserved:
            t.type = self.__class__.reserved[val]
        return t

    def t_newline(self, t):
        r"""[\r\n]+"""
        t.lexer.lineno += t.value.count("\n")

    def t_error(self, t):
        raise SyntaxError("Illegal character '%s'" % t.value[0])

    def p_expression(self, p):
        r"""expression : SELECT projections FROM view where_expression order_by_expression limit_expression offset_expression
                    | boolean_expressions
                    | binary_expr
        """
        if len(p) == 9:
            p[0] = FlatQL(p[2], p[4], p[5], p[6], p[7], p[8])
        else:
            p[0] = p[1]

    def p_projections(self, p):
        r"""projections : projection
                         | projections COMMA projection"""
        self._item_list(p)

    def p_projection(self, p):
        r"""projection : column_aliased
                        | function_call_aliased
                        | binary_expr_aliased
                        | case_aliased
         """
        p[0] = p[1]

    def p_column(self, p):
        r"""column : table PERIOD identifier
         """
        p[0] = Column(p[1], p[3])

    def p_column_aliased(self, p):
        r"""column_aliased : table PERIOD identifier alias
         """
        p[0] = Column(p[1], p[3], p[4])

    def p_case_aliased(self, p):
        r"""case_aliased : case_expr alias
         """
        p[0] = p[1]
        if p[0] is not None:
            p[0].alias = p[2]

    def p_binary_expr_aliased(self, p):
        r"""binary_expr_aliased : binary_expr alias
         """
        p[0] = p[1]
        if p[0] is not None:
            p[0].alias = p[2]

    def p_function_call(self, p):
        r"""function_call : IDENTIFIER LPAREN function_call_args RPAREN"""
        p[0] = FunctionCall(p[1], p[3])

    def p_function_call_aliased(self, p):
        r"""function_call_aliased : IDENTIFIER LPAREN function_call_args RPAREN alias"""
        p[0] = FunctionCall(p[1], p[3], p[5])

    def p_alias(self, p):
        r"""alias : AS identifier
                    | empty
        """
        if len(p) == 3:
            p[0] = p[2]

    def p_function_call_args(self, p):
        r"""function_call_args : function_call_arg
                           | function_call_args COMMA function_call_arg
        """
        if p[1] is not None:
            self._item_list(p)

    def p_function_call_arg(self, p):
        r"""function_call_arg : literal
                           | array_literal
                           | column
                           | table
                           | binary_expr
                           | function_call
                           | case_expr
                           | boolean_expressions
                           | empty
        """
        p[0] = p[1]

    def p_table(self, p):
        r"""table : identifier"""
        p[0] = Table(p[1])

    def p_view(self, p):
        r"""view : identifier"""
        p[0] = View(p[1])

    def p_where_expression(self, p):
        r"""where_expression : where boolean_expressions
                        | empty"""
        if len(p) >= 3:
            p[0] = p[2]

    def p_where(self, p):
        r"""where : WHERE
        """
        p[0] = p[1]

    # def p_binary_expr(self, p):
    #     """ binary_expr : """

    def p_binary_expr_group(self, p):
        """ binary_expr : LPAREN binary_expr RPAREN """
        p[0] = ParenExpression(p[2])

    def p_binary_expr_op(self, p):
        r""" binary_expr : binary_expr PLUS binary_expr
                          | binary_expr MINUS binary_expr
                          | binary_expr TIMES binary_expr
                          | binary_expr DIVIDE binary_expr
                          | binary_expr EXP binary_expr
        """
        p[0] = BinaryExpression(p[1], p[2], p[3])

    def p_binary_expr_operand(self, p):
        """ binary_expr : binary_expr_number
                    | function_call
                    | column
        """
        p[0] = p[1]

    def p_binary_expr_number(self, p):
        """ binary_expr_number : NUMBER
                    | INTEGER
                    | DECIMAL
        """
        p[0] = IntegerLiteral(p[1])

    def p_case(self, p):
        """ case_expr : CASE whens_expr else_expr END
        """
        p[0] = CaseExpression(p[2], p[3])

    def p_whens(self, p):
        """ whens_expr : when_expr
                       | whens_expr when_expr
        """
        self._item_list2(p)

    def p_when(self, p):
        """ when_expr : WHEN boolean_expressions THEN then
        """
        p[0] = WhenExpression(p[2], p[4])

    def p_else(self, p):
        """ else_expr : ELSE else_statement
                    | empty
        """
        if len(p) == 3:
            p[0] = p[2]

    def p_else_statement(self, p):
        """ else_statement : binary_expr
                        | column
                        | function_call
                        | literal
                        | null_expr
                """
        p[0] = p[1]

    def p_then(self, p):
        """ then : binary_expr
                    | column
                    | function_call
                    | literal
                    | null_expr
        """
        p[0] = p[1]

    def p_null_expr(self, p):
        """ null_expr : NULL """
        p[0] = Null()

    def p_boolean_expressions(self, p):
        r"""boolean_expressions : raw_boolean_expressions
        """
        p[0] = LogicalExpression(p[1])

    def p_raw_boolean_expressions(self, p):
        r"""raw_boolean_expressions : boolean_expression
                       | raw_boolean_expressions AND boolean_expression
        """
        self._item_list(p)

    def p_boolean_expression(self, p):
        r"""boolean_expression : comparison_predicate
                       | in_predicate
                       | between_predicate
                       | is_null_predicate
                       | like_predicate
                       | function_call
        """
        p[0] = p[1]

    def p_is_null_predicate(self, p):
        r"""is_null_predicate : predicate_operand IS negated NULL """
        p[0] = IsNull(p[1], p[3] is not None)

    def p_like_predicate(self, p):
        r"""like_predicate : predicate_operand negated LIKE literal """
        p[0] = Like(p[1], p[2] is not None, p[4])

    def p_comparison_predicate(self, p):
        r"""comparison_predicate : predicate_operand comparison_operator comparison_rhs"""
        p[0] = Comparison(p[1], p[2], p[3])

    def p_comparison_rhs(self, p):
        r"""comparison_rhs : literal
                            | function_call
                            | column
        """
        p[0] = p[1]

    def p_in_predicate(self, p):
        r"""in_predicate : predicate_operand negated IN LPAREN literals RPAREN """
        p[0] = In(p[1], p[2] is not None, p[5])

    def p_between_predicate(self, p):
        r"""between_predicate : predicate_operand negated BETWEEN literal AND literal """
        p[0] = Between(p[1], p[2] is not None, p[4], p[6])

    def p_predicate_operand(self, p):
        r"""predicate_operand : column
                               | function_call
        """
        p[0] = p[1]

    def p_order_by_expression(self, p):
        r"""order_by_expression : ORDER BY order_by_items_expression
                                | empty
        """
        if len(p) == 4:
            p[0] = p[3]

    def p_order_by_items_expression(self, p):
        r"""order_by_items_expression : order_by_item
                                | order_by_items_expression COMMA order_by_item
        """
        self._item_list(p)

    def p_order_by_item(self, p):
        r"""order_by_item : identifier order_by_direction
        """
        p[0] = OrderByItem(p[1], p[2])

    def p_order_by_direction(self, p):
        r"""order_by_direction : DESC
                                | ASC
                                | empty
        """
        p[0] = p[1]

    def p_offset_expression(self, p):
        r"""offset_expression : OFFSET INTEGER
                            | empty
        """
        if p[1] is not None:
            p[0] = Offset(IntegerLiteral(p[2]))

    def p_limit_expression(self, p):
        r"""limit_expression : LIMIT INTEGER
                            | empty
        """
        if p[1] is not None:
            p[0] = Limit(IntegerLiteral(p[2]))

    def p_negated(self, p):
        r"""negated : NOT
                | empty
        """
        p[0] = p[1]

    def p_comparison_operator(self, p):
        r"""comparison_operator : EQ
                                | NE
                                | LT
                                | LE
                                | GT
                                | GE
        """
        p[0] = p[1]

    def p_identifier(self, p):
        r"""identifier : IDENTIFIER
                        | QUOTED_IDENTIFIER
        """
        p[0] = Identifier(p[1], p.slice[1].type == 'QUOTED_IDENTIFIER')

    def p_literals(self, p):
        r""" literals : literal
                      | literals COMMA literal
        """
        self._item_list(p)

    def p_literal(self, p):
        r""" literal : NUMBER
                    | STRING
                    | INTEGER
                    | DECIMAL
        """
        p[0] = Literal(p[1], p.slice[1].type)

    def p_array_literal(self, p):
        r""" array_literal : LBRACK literals RBRACK
        """
        p[0] = ArrayLiteral(p[2])

    def p_empty(self, p):
        """empty :"""
        pass

    def p_error(self, p):
        if p:
            err = SyntaxError()
            err.lineno = p.lineno
            err.text = p.lexer.lexdata
            err.token_value = p.value

            text_lines = err.text.split("\n")
            line_lengths = [len(line) + 1 for line in text_lines]
            err_line_offset = sum(line_lengths[:err.lineno - 1])

            err.line = text_lines[err.lineno - 1]
            err.offset = p.lexpos - err_line_offset
            err.msg = ("Syntax error at position %d (%s)" % (err.offset, str(err.token_value)))

            def _print_error(parser):
                pointer = " " * parser.literal + "^" * len(parser.token_value)
                print(parser.line + "\n" + pointer)

            _print_error = types.MethodType(_print_error, err)
            err.print_file_and_line = _print_error

            raise err
        raise SyntaxError("Syntax error in input. Check your statement")

    # noinspection PyUnresolvedReferences
    def parse(self, flatql: str):
        return yacc.parse(flatql)

    def _item_list(self, p):
        if len(p) == 2:
            p[0] = [p[1]]
        elif isinstance(p[1], list):
            p[1].append(p[3])
            p[0] = p[1]
        else:
            p[0] = None

    def _item_list2(self, p):
        if len(p) == 2:
            p[0] = [p[1]]
        elif isinstance(p[1], list):
            p[1].append(p[2])
            p[0] = p[1]
        else:
            p[0] = None


def parse_flatql(flatql: str) -> FlatQL:
    return FlatQLParser().parse(flatql)


def parse_join_condition(join_on: str) -> List[Predicate]:
    join_condition = FlatQLParser().parse(join_on)
    if isinstance(join_condition, LogicalExpression):
        for item in join_condition:
            if not isinstance(item, Predicate):
                raise SyntaxError("Join condition must be logic expression")
    else:
        raise SyntaxError("Join condition must be logic expression")
    return join_condition


def parse_binary_expression(binary_expr: str) -> Union[BinaryExpression, FunctionCall]:
    binary_expression = FlatQLParser().parse(binary_expr)
    if (not isinstance(binary_expression, BinaryExpression)) and \
            (not isinstance(binary_expression, FunctionCall)):
        raise SyntaxError("Invalid binary expression")
    return binary_expression


def parse_qualified_name(raw_name: str) -> QualifiedName:
    if re.fullmatch(QUALIFIED_NAME_PATTERN, raw_name):
        return QualifiedName(*[Identifier(n) for n in raw_name.split(".")])
    raise SyntaxError(f"Invalid qualified name: '{raw_name}'")
