from typing import List, Optional

from .function_call_rewriter import FunctionCallRewriterFactory
from .sql import SqlAggFunctionCall, SqlIdentifier, SqlFunctionCall, SqlLiteral, \
    SqlAdditiveAggrFunctionCall, SqlLogicalExpression, SqlCaseExpression, SqlWhenExpression, SqlIsNull
from .sql_generic import SqlNode
from ..parser.ast import IntegerLiteral


class ClickhouseSqlRewriterFactory(FunctionCallRewriterFactory):

    def avg(self, args: List[SqlNode], alias: Optional[SqlIdentifier] = None) -> SqlNode:
        avg = SqlAggFunctionCall("AVG", args)
        return self._if(self._is_nan(avg), SqlLiteral(IntegerLiteral(0)), avg, alias=alias)

    def avg_if(self, args: List[SqlNode], alias: Optional[SqlIdentifier] = None) -> SqlNode:
        if len(args) != 2 or (not isinstance(args[1], SqlLogicalExpression)):
            raise SyntaxError("Illegal function call for avg_if, for example: avg_if([column], [logical expression])")
        avg = SqlAggFunctionCall("avgIf", args)
        return self._if(self._is_nan(avg), SqlLiteral(IntegerLiteral(0)), avg, alias=alias)

    def count_if(self, args: List[SqlNode], alias: Optional[SqlIdentifier] = None) -> SqlNode:
        if len(args) != 2 or (not isinstance(args[1], SqlLogicalExpression)):
            raise SyntaxError(
                "Illegal function call for count_if, for example: count_if([column], [logical expression])")
        return SqlAdditiveAggrFunctionCall("countIf", args, alias=alias)

    def max_if(self, args: List[SqlNode], alias: Optional[SqlIdentifier] = None) -> SqlNode:
        if len(args) != 2 or (not isinstance(args[1], SqlLogicalExpression)):
            raise SyntaxError("Illegal function call for max_if, for example: max_if([column], [logical expression])")
        return SqlAggFunctionCall("maxIf", args, alias=alias)

    def min_if(self, args: List[SqlNode], alias: Optional[SqlIdentifier] = None) -> SqlNode:
        if len(args) != 2 or (not isinstance(args[1], SqlLogicalExpression)):
            raise SyntaxError("Illegal function call for min_if, for example: min_if([column], [logical expression])")
        return SqlAggFunctionCall("minIf", args, alias=alias)

    def sum_if(self, args: List[SqlNode], alias: Optional[SqlIdentifier] = None) -> SqlNode:
        if len(args) != 2 or (not isinstance(args[1], SqlLogicalExpression)):
            raise SyntaxError("Illegal function call for sum_if, for example: sum_if([column], [logical expression])")
        return SqlAdditiveAggrFunctionCall("sumIf", args, alias=alias)

    def to_year(self, args: List[SqlNode], alias: Optional[SqlIdentifier] = None) -> SqlNode:
        return SqlFunctionCall("toYear", args, alias=alias)

    def to_quarter(self, args: List[SqlNode], alias: Optional[SqlIdentifier] = None) -> SqlNode:
        return SqlFunctionCall("toQuarter", args, alias=alias)

    def to_month(self, args: List[SqlNode], alias: Optional[SqlIdentifier] = None) -> SqlNode:
        return SqlFunctionCall("toMonth", args, alias=alias)

    def to_week(self, args: List[SqlNode], alias: Optional[SqlIdentifier] = None) -> SqlNode:
        return SqlFunctionCall("toWeek", args, alias=alias)

    def date_format(self, args: List[SqlNode], alias: Optional[SqlIdentifier] = None) -> SqlNode:
        return SqlFunctionCall("formatDateTime", args, alias=alias)

    def if_null_or_empty(self, args: List[SqlNode], alias: Optional[SqlIdentifier] = None) -> SqlNode:
        if len(args) != 2:
            raise SyntaxError("Illegal function call for if_null_or_empty, for example: "
                              "if_null_or_empty([column], [replacement literal])")
        when_expression = SqlLogicalExpression([SqlWhenExpression(self.is_empty([args[0]]), args[1])])
        return SqlCaseExpression(when_expression, args[0], alias=alias)

    def if_null(self, args: List[SqlNode], alias: Optional[SqlIdentifier] = None) -> SqlNode:
        if len(args) != 2:
            raise SyntaxError("Illegal function call for if_null, for example: "
                              "if_null([column], [replacement literal])")
        when_expression = SqlLogicalExpression([SqlWhenExpression(SqlIsNull(args[0]), args[1])])
        return SqlCaseExpression(when_expression, args[0], alias=alias)

    def round(self, args: List[SqlNode], alias: Optional[SqlIdentifier] = None) -> SqlNode:
        return SqlFunctionCall("round", args, alias=alias)

    def has(self, args: List[SqlNode], alias: Optional[SqlIdentifier] = None) -> SqlNode:
        return SqlFunctionCall("has", args, alias=alias)

    def has_all(self, args: List[SqlNode], alias: Optional[SqlIdentifier] = None) -> SqlNode:
        return SqlFunctionCall("hasAll", args, alias=alias)

    def has_any(self, args: List[SqlNode], alias: Optional[SqlIdentifier] = None) -> SqlNode:
        return SqlFunctionCall("hasAny", args, alias=alias)

    def is_empty(self, args: List[SqlNode], alias: Optional[SqlIdentifier] = None) -> SqlNode:
        return SqlFunctionCall("empty", args, alias=alias)

    def now(self, args: List[SqlNode], alias: Optional[SqlIdentifier] = None) -> SqlNode:
        return SqlFunctionCall("now", args, alias=alias)

    def _if(self, cond: SqlNode, then: SqlNode, else_: SqlNode, alias: Optional[SqlIdentifier] = None) -> SqlNode:
        return SqlFunctionCall("if", [cond, then, else_], alias=alias)

    def _is_nan(self, expr: SqlNode) -> SqlNode:
        return SqlFunctionCall("isNaN", [expr])
