from abc import ABC
from typing import Callable, List

from .sql import SqlIdentifier, SqlCountDistinctFunctionCall, SqlTable, SqlCountFunctionCall, SqlSumFunctionCall, \
    SqlAggFunctionCall, SqlFunctionCall
from .sql_generic import SqlNode, SqlKeyword

__sum__, __count__, __sum_if__, __count_if__, __count_distinct__ = "SUM", "COUNT", "SUM_IF", "COUNT_IF", "COUNT_DISTINCT"
__aggr_function_call__ = ["AVG", "MIN", "MAX", "AVG_IF", "MIN_IF", "MAX_IF"]
__additive_aggr_function_call__ = [__sum__, __count__, __sum_if__, __count_if__, __count_distinct__]
__predicate_function_call__ = ["HAS", "HAS_ALL", "HAS_ANY", "IS_EMPTY"]
__sql_functions__ = ["TO_YEAR", "TO_QUARTER", "TO_MONTH", "TO_WEEK", "DATE_FORMAT", "NOW", "IF_NULL_OR_EMPTY",
                     "IF_NULL", "ROUND", "TO_STRING", "TO_INTEGER", "TO_FLOAT", "TO_DATE", "TO_DATETIME"
                     ] + __predicate_function_call__ + __aggr_function_call__ + __additive_aggr_function_call__


class FunctionCallRewriterFactory(ABC):

    def get_rewriter(self, name: str) -> Callable:
        if not is_function_call_supported(name.upper()):
            raise RuntimeError(f"Unsupported function: {name}")
        elif hasattr(self, name.lower()):
            return getattr(self, name.lower())
        return None


def is_additive_aggr_function_call(function_name: str):
    return function_name in __additive_aggr_function_call__


def is_aggr_function_call(function_name: str):
    return function_name in __aggr_function_call__ or \
           function_name in __additive_aggr_function_call__


def is_function_call_supported(function_name: str):
    return function_name in __sql_functions__


def rewrite_sql_function_call(function_name, args: List[SqlNode], alias: SqlIdentifier):
    if function_name.upper() == __count_distinct__:
        return SqlCountDistinctFunctionCall(args, alias=alias)
    elif function_name.upper() in [__count__, __count_if__]:
        if len(args) == 1 and isinstance(args[0], SqlTable):
            return SqlCountFunctionCall([SqlKeyword("*")], alias=alias)
        return SqlCountFunctionCall(args, alias=alias)
    elif function_name.upper() in [__sum__, __sum_if__]:
        return SqlSumFunctionCall(args, alias=alias)
    elif is_aggr_function_call(function_name.upper()):
        return SqlAggFunctionCall(function_name.upper(), args, alias=alias)
    else:
        return SqlFunctionCall(function_name.upper(), args, alias=alias)
