from abc import ABC, ABCMeta, abstractmethod
from typing import List, Optional, Dict

from .algorithm import detect_involved_tables
from .frozen import immutable_property
from .select import Select
from .sql import SqlNode, SqlContext, SqlPredicate, SqlOrderByItem, SqlLimit, SqlOffset, \
    SqlKeyword, SqlAggFunctionCall, K_INNER_JOIN, SqlFunctionCall, SqlSumFunctionCall, \
    K_LEFT_JOIN, Subquery, SqlComparison, SqlColumn, O_EQ, SqlLogicalExpression, RewritableSqlNode, \
    K_CROSS_JOIN, SqlIdentifier, SqlTable, SqlAdditiveAggrFunctionCall, SqlBinaryExpression, SqlParenExpression, \
    SqlCaseExpression
from .sql_generic import SqlContextWrapper, SqlNodeIntrospection


class SqlPredicateAppender(metaclass=ABCMeta):

    @abstractmethod
    def append(self, introspection: SqlNodeIntrospection, table: SqlTable): pass


class QueryNodeParameter:

    def __init__(self,
                 projections: Optional[List[SqlNode]] = None,
                 predicates: Optional[SqlLogicalExpression] = None,
                 group_bys: Optional[List[SqlNode]] = None,
                 order_by_items: Optional[List[SqlNode]] = None,
                 limit: Optional[SqlLimit] = None,
                 offset: Optional[SqlOffset] = None):
        self._projections = projections if projections is not None else list()
        self._predicates = predicates if predicates is not None else SqlLogicalExpression([])
        self._group_bys = group_bys if group_bys is not None else list()
        self._order_by_items = order_by_items if order_by_items is not None else list()
        self.limit = limit
        self.offset = offset

    @immutable_property
    def projections(self):
        return self._projections

    @property
    def predicates(self) -> SqlNode:
        return self._predicates

    @immutable_property
    def group_bys(self):
        return self._group_bys

    @immutable_property
    def order_by_items(self):
        return self._order_by_items

    def append_projections(self, projection: SqlNode):
        if projection not in self._projections:
            self._projections.append(projection)

    def append_group_by_item(self, item: SqlNode):
        if item not in self._group_bys:
            self._group_bys.append(item)

    def append_predicate(self, predicate: SqlNode):
        if predicate not in self._predicates:
            self._predicates.append(predicate)

    def append_predicate_list(self, predicates: List[SqlNode]):
        for predicate in predicates:
            self.append_predicate(predicate)


class QueryNode(SqlNode, ABC):

    # noinspection PyTypeChecker
    def __init__(self, query_parameter: QueryNodeParameter):
        self.query_parameter = query_parameter
        self.left: QueryNode = None
        self.right: QueryNode = None
        self._children_joiner = K_INNER_JOIN
        self.sql_node_reference: Dict[SqlNode, SqlNode] = {}

    @immutable_property
    def references(self) -> Dict[SqlNode, SqlNode]:
        return self.sql_node_reference

    def to_sql(self, context: 'SqlContext') -> str:
        return self._create_query(context).to_sql(context)

    def _create_query(self, context: 'SqlContext') -> Select: pass

    def get_reference(self, sql_node: SqlNode) -> SqlNode:
        return self.sql_node_reference.get(sql_node)


class RootableQueryNode(QueryNode, ABC):

    def __init__(self, query_parameter: QueryNodeParameter, is_root: Optional[bool] = False):
        super().__init__(query_parameter)
        self._is_root = is_root

    def _fill_root_query_parameters(self, select: Select, query_parameter: Optional[QueryNodeParameter] = None):
        if query_parameter is None:
            query_parameter = self.query_parameter

        select.limit(query_parameter.limit) \
            .offset(query_parameter.offset) \
            .order_by_list(query_parameter.order_by_items)

    def _create_reference_column(self, query: Select, reference: Dict[SqlNode, SqlNode]):
        return {n: SqlColumn(query, r) for n, r in reference.items()}


class SimpleQueryNode(RootableQueryNode):

    def __init__(self, query_parameter: QueryNodeParameter, is_root: Optional[bool] = False):
        super().__init__(query_parameter)
        self._is_root = is_root

    def _create_query(self, context: 'SqlContext') -> Select:
        self.query_parameter = self._rewrite_query_parameter(context, self.query_parameter)

        introspection = SqlNodeIntrospection.inspect(context, [
            self.query_parameter.projections,
            self.query_parameter.predicates
        ])
        select = Select()

        if introspection.has_multiple_tables():
            tables_name = introspection.tables_name
            main_table = context.get_main_table(tables_name)
            self._build_join(context, introspection, select, main_table)
        else:
            select.from_(introspection.table)

        self._build_projections(context, select)
        self._build_predicates(introspection, select)
        self._build_group_by(select)
        self._fill_root_query_parameters(select)

        return select

    def _rewrite_query_parameter(self, context: SqlContext, query_parameter: QueryNodeParameter):
        return query_parameter

    def _build_projections(self, context: SqlContext, select: Select):
        for projection in self.query_parameter.projections:
            if hasattr(projection, "alias"):
                if not self._is_root:
                    projection.alias = context.alias(projection)
                    self.sql_node_reference[projection] = projection.alias
            select.project(projection)

    def _build_group_by(self, select: Select):
        if self.query_parameter.group_bys is not None:
            for item in self.query_parameter.group_bys:
                if isinstance(item, SqlIdentifier):
                    select.group_by(item)
                else:
                    if item.alias is not None:
                        select.group_by(item.alias)
                    else:
                        select.group_by(item)

    def _build_predicates(self, introspection: SqlNodeIntrospection, select: Select):
        if self.query_parameter.predicates is not None:
            for predicate in self.query_parameter.predicates:
                if isinstance(predicate, SqlPredicate):
                    if isinstance(predicate.operand, SqlAggFunctionCall):
                        select.having(predicate)
                    else:
                        select.predicate(predicate)
                elif isinstance(predicate, SqlFunctionCall):
                    if not introspection.get_node_introspection(predicate).has_aggr_function_calls():
                        select.predicate(predicate)
                else:
                    raise RuntimeError(f"Invalid predicate : {predicate}")

    def _build_join(self, context: SqlContext, introspection: SqlNodeIntrospection, select: Select, main_table: str):
        tables_dict = introspection.tables_dict
        tables_name = introspection.tables_name
        relation_graph = context.get_relation_graph()

        join_order = list()
        parent_tree, _ = detect_involved_tables(main_table, tables_name, {main_table}, relation_graph,
                                                join_order=join_order)
        for fk_table_name in join_order:
            parent_table_name = parent_tree[fk_table_name]
            self._append_join(context, tables_dict, select, fk_table_name, parent_table_name)

        select.from_(tables_dict.get(main_table))

    def _append_join(self, context: SqlContext, tables_dict: Dict[str, SqlTable],
                     select: Select, fk_table_name: str, parent_table_name: str):
        table_schema = context.get_table_schema(parent_table_name)

        if table_schema.has_join(fk_table_name):
            join = table_schema.get_join(fk_table_name)
            join_condition = self._get_join_on_expr(context, parent_table_name, join.target, tables_dict)
            select.join(SqlKeyword(join.join_type), tables_dict.get(join.target), join_condition)
        else:
            join = context.get_table_schema(fk_table_name).get_join(parent_table_name)
            join_condition = self._get_join_on_expr(context, join.target, fk_table_name, tables_dict)
            select.join(SqlKeyword(join.join_type), tables_dict.get(fk_table_name), join_condition)

    def _get_join_on_expr(self, context: SqlContext, parent_table_name: str, table_name: str, tables_dict: Dict[str, SqlTable]):
        return context.get_join_condition(parent_table_name, table_name)


class QueryCombineNode(RootableQueryNode):

    def __init__(self,
                 introspection: SqlNodeIntrospection,
                 order_by_items: Optional[List[SqlNode]] = None,
                 limit: Optional[SqlNode] = None,
                 offset: Optional[SqlNode] = None,
                 is_root: Optional[bool] = False):
        super().__init__(QueryNodeParameter(
            order_by_items=order_by_items,
            limit=limit,
            offset=offset
        ), is_root)
        self._introspection = introspection

    def _create_query(self, context: 'SqlContext') -> Select:
        if self.right is None:
            query = self.left._create_query(context)
            if self._is_root:
                query = self._build_query(query, self.left)
            else:
                self.sql_node_reference = self.left.sql_node_reference
        else:
            query = self._build_binary_query(context)

        return query

    # noinspection PyUnresolvedReferences
    def _rewrite_projection(self, projection: SqlNode, ref_columns_dict: Dict[SqlNode, SqlNode]):
        items = sorted(ref_columns_dict.items(), key=lambda d: d[1])
        if isinstance(projection, SqlBinaryExpression) or \
                isinstance(projection, SqlParenExpression) or \
                isinstance(projection, SqlCaseExpression) or \
                (isinstance(projection, SqlFunctionCall) and not isinstance(projection, SqlAggFunctionCall)):
            is_whole_rewrite = self._introspection.is_whole_rewrite(projection)
            if is_whole_rewrite:
                reference = ref_columns_dict.get(projection)
                projection = projection.copy()
                projection.rewrite_reference(projection, reference, is_whole_rewrite=True)
            else:
                projection = projection.copy()
                for original, reference in items:
                    projection.rewrite_reference(original, reference)
        else:
            if isinstance(projection, RewritableSqlNode):
                reference = ref_columns_dict.get(projection)
                projection = projection.copy()
                if reference is not None:
                    projection.rewrite_reference(projection, reference)

        return projection

    def _build_query(self, query: Select, node: QueryNode) -> Select:
        select = Select()
        ref_columns_dict = self._create_reference_column(query, node.references)
        for projection in self._introspection.projections:
            select.project(self._rewrite_projection(projection, ref_columns_dict))
        select.from_(Subquery(query))
        self._fill_root_query_parameters(select)
        return select

    def _build_binary_query(self, context: 'SqlContext') -> Select:
        select = Select()
        left_query = self.left._create_query(context)
        right_query = self.right._create_query(context)

        select.from_(Subquery(left_query))
        self._build_projections(left_query, right_query, select, context)
        self._build_join(left_query, right_query, select)

        if self._is_root:
            self._fill_root_query_parameters(select)

        return select

    def _build_join(self, left_query: QueryNode, right_query: QueryNode, select: Select):
        if self._introspection.has_dimensions():
            comparisons = []
            for dimension in self._introspection.dimensions:
                left_column = SqlColumn(left_query, self.left.get_reference(dimension))
                right_column = SqlColumn(right_query, self.right.get_reference(dimension))
                comparisons.append(SqlComparison(left_column, O_EQ, right_column))
            select.join(K_INNER_JOIN, Subquery(right_query), SqlLogicalExpression(comparisons))
        else:
            select.join(K_CROSS_JOIN, Subquery(right_query))

    def _build_projections(self, left_query: QueryNode, right_query: QueryNode, select: Select, context: 'SqlContext'):
        left_ref_columns_dict = self._create_reference_column(left_query, self.left.references)
        right_ref_columns_dict = self._create_reference_column(right_query, self.right.references)
        ref_columns_dict = {**left_ref_columns_dict, **right_ref_columns_dict}

        for projection in self._introspection.projections:
            if self._is_root:
                select.project(self._rewrite_projection(projection, ref_columns_dict))
            else:
                reference = ref_columns_dict.get(projection)
                if reference is not None:
                    reference.alias = context.alias(reference)
                    self.sql_node_reference[projection] = reference.alias
                    select.project(reference)

    def to_sql(self, context: 'SqlContext') -> str:
        return super().to_sql(SqlContextWrapper(context))


class AdditiveAggrQueryCombineNode(QueryNode):

    def __init__(self, introspection: SqlNodeIntrospection, aggr_function_calls: List[SqlFunctionCall],
                 function_call_primary_key: SqlColumn):
        super().__init__(QueryNodeParameter())
        self._introspection = introspection
        self._aggr_function_calls = aggr_function_calls
        self._function_call_primary_key = function_call_primary_key
        self.children_joiner = K_LEFT_JOIN

    def _create_query(self, context: 'SqlContext') -> Select:
        select = Select()
        dimension_query = self.left._create_query(context)
        aggr_function_call_query = self.right._create_query(context)

        select.from_(Subquery(dimension_query))

        self._build_dimension_projection(context, select, dimension_query)
        self._build_aggr_function_call_projection(context, select, aggr_function_call_query)

        dimension_primary_key_column = SqlColumn(dimension_query,
                                                 self.left.get_reference(self._function_call_primary_key))
        aggr_function_call_primary_key_column = SqlColumn(aggr_function_call_query,
                                                          self.right.get_reference(self._function_call_primary_key))
        select.join(self.children_joiner, Subquery(aggr_function_call_query),
                    SqlComparison(dimension_primary_key_column, O_EQ, aggr_function_call_primary_key_column))

        return select

    def _build_dimension_projection(self, context: SqlContext, select: Select, dimension_query: Select):
        for dimension in self._introspection.dimensions:
            dimension_column = SqlColumn(dimension_query, self.left.get_reference(dimension))
            dimension_column.alias = context.alias(dimension)
            select.project(dimension_column)
            select.group_by(dimension_column.alias)

            self.sql_node_reference[dimension] = dimension_column.alias

    def _build_aggr_function_call_projection(self, context: SqlContext, select: Select,
                                             aggr_function_call_query: Select):
        for aggr_function_call in self._aggr_function_calls:
            aggr_function_call_column = SqlColumn(aggr_function_call_query,
                                                  self.right.get_reference(aggr_function_call))
            if isinstance(aggr_function_call, SqlAdditiveAggrFunctionCall):
                function_call_wrapper = SqlSumFunctionCall([aggr_function_call_column])
            else:
                function_call_wrapper = SqlAggFunctionCall(
                    aggr_function_call.function_name, [aggr_function_call_column])

            function_call_wrapper.alias = context.alias(function_call_wrapper)
            select.project(function_call_wrapper)

            self.sql_node_reference[aggr_function_call] = function_call_wrapper.alias

    def to_sql(self, context: 'SqlContext') -> str:
        return super().to_sql(SqlContextWrapper(context))


class DimensionQueryNode(SimpleQueryNode):

    def __init__(self, introspection: SqlNodeIntrospection, function_call_primary_key: SqlColumn):
        super().__init__(QueryNodeParameter())
        self._introspection = introspection
        self._function_call_primary_key = function_call_primary_key

    def _create_query(self, context: 'SqlContext') -> Select:
        introspection = SqlNodeIntrospection.inspect(context, [
            [self._function_call_primary_key],
            self._introspection.dimensions,
            _flatten(self._introspection.where_predicates.values())
        ])
        select = Select()

        self._build_projections_and_group_by(context, select)

        if introspection.has_multiple_tables():
            tables_name = introspection.tables_name
            main_table_name = context.get_main_table(tables_name)
            self._build_join(context, introspection, select, main_table_name)
            main_table = self._introspection.tables_dict[main_table_name]
            predicates = self._introspection.where_predicates.get(main_table)
            if predicates is not None:
                for predicate in predicates:
                    select.predicate(predicate)
        else:
            select.from_(introspection.table)
            for predicates in introspection.where_predicates.values():
                for predicate in predicates:
                    select.predicate(predicate)
        return select

    def _build_projections_and_group_by(self, context: SqlContext, select: Select):
        self._function_call_primary_key.alias = context.alias(self._function_call_primary_key)

        select.project(self._function_call_primary_key)
        select.group_by(self._function_call_primary_key.alias)
        self.sql_node_reference[self._function_call_primary_key] = self._function_call_primary_key.alias

        for dimension in self._introspection.dimensions:
            dimension.alias = context.alias(dimension)
            select.project(dimension)
            select.group_by(dimension.alias)

            self.sql_node_reference[dimension] = dimension.alias

    # noinspection PyUnresolvedReferences
    def _get_join_on_expr(self, context: SqlContext, parent_table_name: str, table_name: str, tables_dict: Dict[str, SqlTable]):
        join_on_expr = super()._get_join_on_expr(context, parent_table_name, table_name, tables_dict)
        predicates = self._introspection.where_predicates.get(tables_dict[table_name])
        if predicates is not None:
            for predicate in predicates:
                if predicate is not predicates:
                    join_on_expr.append(predicate)
        return join_on_expr

    def to_sql(self, context: 'SqlContext') -> str:
        return super().to_sql(SqlContextWrapper(context))


class AdditiveAggrQueryNode(SimpleQueryNode):

    def __init__(self, introspection: SqlNodeIntrospection, additive_aggr_function_calls: List[SqlFunctionCall],
                 function_call_primary_key: Optional[SqlColumn] = None):
        super().__init__(QueryNodeParameter())
        self._introspection = introspection
        self._additive_aggr_function_calls = additive_aggr_function_calls
        self._function_call_primary_key = function_call_primary_key

    def _rewrite_query_parameter(self, context: SqlContext, query_parameter: QueryNodeParameter):
        if self._function_call_primary_key is not None:
            self._function_call_primary_key.alias = context.alias(self._function_call_primary_key)
            query_parameter.append_projections(self._function_call_primary_key)
            query_parameter.append_group_by_item(self._function_call_primary_key.alias)

            self.sql_node_reference[self._function_call_primary_key] = self._function_call_primary_key.alias

        for additive_aggr_function_call in self._additive_aggr_function_calls:
            additive_aggr_function_call.alias = context.alias(additive_aggr_function_call)
            query_parameter.append_projections(additive_aggr_function_call)

            self.sql_node_reference[additive_aggr_function_call] = additive_aggr_function_call.alias

        introspection = SqlNodeIntrospection.inspect(context, [self._additive_aggr_function_calls])
        predicates = self._introspection.get_predicates([introspection.table])
        query_parameter.append_predicate_list(predicates)

        return query_parameter

    def to_sql(self, context: 'SqlContext') -> str:
        return super().to_sql(SqlContextWrapper(context))


class SimpleAggrQueryNode(SimpleQueryNode):

    def __init__(self, introspection: SqlNodeIntrospection,
                 aggr_function_calls: List[SqlAggFunctionCall], dimensions: Optional[List[SqlNode]] = None):
        super().__init__(QueryNodeParameter())
        self._aggr_function_calls = aggr_function_calls
        self._introspection = introspection
        self._dimensions = dimensions if dimensions is not None else []

    def _rewrite_query_parameter(self, context: SqlContext, query_parameter: QueryNodeParameter):
        for projection in (self._dimensions + self._aggr_function_calls):
            if hasattr(projection, "alias"):
                query_parameter.append_projections(projection)
                projection.alias = context.alias(projection)
                self.sql_node_reference[projection] = projection.alias

        for group_by_item in self._dimensions:
            if hasattr(group_by_item, "alias"):
                query_parameter.append_group_by_item(group_by_item.alias)

        for predicates in self._introspection.where_predicates.values():
            for predicate in predicates:
                query_parameter.append_predicate(predicate)

        return query_parameter

    def to_sql(self, context: 'SqlContext') -> str:
        return super().to_sql(SqlContextWrapper(context))


class QueryTree(SqlNode):

    def __init__(self, order_by_items: List[SqlNode],
                 limit: SqlLimit,
                 offset: SqlOffset):
        self._order_by_items = order_by_items
        self._limit = limit
        self._offset = offset
        self._root = None

    def build_tree_nodes(self, context: SqlContext, introspection: SqlNodeIntrospection):
        if introspection.has_additive_aggr_function_calls():
            if introspection.has_multiple_tables():
                self._root = self._build_additive_aggr_query_tree(context, introspection)
            else:
                self._root = self._create_simple_root(introspection)
        else:
            self._root = self._create_simple_root(introspection)

    def _build_additive_aggr_query_tree(self, context: SqlContext,
                                        introspection: SqlNodeIntrospection,
                                        is_root: Optional[bool] = True) -> QueryNode:
        aggr_combine_nodes = self._create_additive_aggr_combine_nodes(context, introspection)

        root = self._create_root_combine_query_node(introspection, is_root)
        parent = root
        for aggr_combine_node in aggr_combine_nodes:
            parent = self._append_combine_query_node(introspection, parent, aggr_combine_node)

        return root

    def _append_combine_query_node(self, introspection: SqlNodeIntrospection,
                                   parent: QueryNode, aggr_combine_node: QueryNode) -> QueryNode:
        if parent.left is None:
            parent.left = aggr_combine_node
        elif parent.right is None:
            parent.right = aggr_combine_node
        else:
            temp_parent = QueryCombineNode(introspection)
            temp_parent.left = parent.left
            temp_parent.right = aggr_combine_node
            parent.left = temp_parent
            parent = temp_parent
        return parent

    def _create_root_combine_query_node(self, introspection: SqlNodeIntrospection,
                                        is_root: Optional[bool] = True) -> QueryNode:
        return QueryCombineNode(
            introspection,
            order_by_items=self._order_by_items,
            limit=self._limit,
            offset=self._offset,
            is_root=is_root
        )

    def _create_additive_aggr_combine_nodes(self, context: SqlContext,
                                            introspection: SqlNodeIntrospection) -> List[QueryNode]:
        combine_nodes = list()
        aggr_function_calls_dict = introspection.aggr_function_calls
        aggr_function_calls_table = sorted(list(aggr_function_calls_dict.keys()))
        for table in aggr_function_calls_table:
            aggr_function_calls = aggr_function_calls_dict.get(table)
            if introspection.has_additive_aggr_function_calls(aggr_function_calls):
                primary_key = context.create_table_primary_key(table)
                combine_node = AdditiveAggrQueryCombineNode(introspection, aggr_function_calls, primary_key)
                combine_node.left = DimensionQueryNode(introspection, primary_key)
                combine_node.right = AdditiveAggrQueryNode(introspection, aggr_function_calls, primary_key)

                combine_nodes.append(combine_node)
            else:
                combine_nodes.append(SimpleAggrQueryNode(introspection, aggr_function_calls, introspection.dimensions))

        return combine_nodes

    def _create_simple_root(self, introspection: SqlNodeIntrospection) -> QueryNode:
        return self._create_simple_node(
            projections=introspection.projections,
            predicates=introspection.predicates,
            group_bys=(introspection.dimensions if introspection.has_aggr_function_calls() else []),
            is_root=True
        )

    def _create_simple_node(self, projections: Optional[List[SqlNode]] = None,
                            predicates: Optional[List[SqlNode]] = None,
                            group_bys: Optional[List[SqlNode]] = None,
                            is_root: Optional[bool] = False) -> QueryNode:
        return SimpleQueryNode(
            query_parameter=QueryNodeParameter(
                projections=projections,
                predicates=predicates,
                group_bys=group_bys,
                order_by_items=self._order_by_items,
                limit=self._limit,
                offset=self._offset
            ),
            is_root=is_root
        )

    def to_sql(self, context: 'SqlContext') -> str:
        return self._root.to_sql(context)


def create_query_tree(context: SqlContext, introspection: SqlNodeIntrospection,
                      order_by_items: Optional[List[SqlOrderByItem]] = None,
                      limit: Optional[SqlLimit] = None, offset: Optional[SqlOffset] = None,
                      sql_predicate_appender: Optional[SqlPredicateAppender] = None):
    order_by_items = order_by_items if order_by_items is not None else []

    if sql_predicate_appender is not None:
        tables = introspection.tables
        for table in tables:
            sql_predicate_appender.append(introspection, table)

    query_tree = QueryTree(order_by_items, limit, offset)
    query_tree.build_tree_nodes(context, introspection)

    return query_tree


def _flatten(xss):
    return [x for xs in xss for x in xs]
