from typing import List, Optional

from .sql import SqlNode, SqlContext, SqlPredicate, SqlJoin, SqlKeyword


class Select(SqlNode):

    def __init__(self):
        self._projections = list()
        self._predicates = list()
        self._joins = list()
        self._groups = list()
        self._having_predicates = list()
        self._order_by_items = list()
        self._limit = None
        self._offset = None

    def project(self, node: SqlNode) -> 'Select':
        if node is not None and node not in self._projections:
            self._projections.append(node)
        return self

    def from_(self, table_reference: SqlNode) -> 'Select':
        self.table_reference = table_reference
        return self

    def predicate(self, predicate: SqlPredicate) -> 'Select':
        if predicate is not None:
            self._predicates.append(predicate)
        return self

    def having(self, predicate: SqlPredicate) -> 'Select':
        if predicate is not None:
            self._having_predicates.append(predicate)
        return self

    def group_by(self, node: SqlNode) -> 'Select':
        if node is not None:
            self._groups.append(node)
        return self

    def order_by(self, node: SqlNode) -> 'Select':
        if node is not None:
            self._order_by_items.append(node)
        return self

    def order_by_list(self, nodes: List[SqlNode]) -> 'Select':
        for node in nodes:
            self.order_by(node)
        return self

    def limit(self, limit: SqlNode) -> 'Select':
        self._limit = limit
        return self

    def offset(self, offset: SqlNode) -> 'Select':
        self._offset = offset
        return self

    def join(self, join_type: SqlKeyword, subclause: SqlNode, on_expr: Optional[SqlNode] = None) -> 'Select':
        self._joins.append(SqlJoin(join_type, subclause, on_expr))
        return self

    def to_sql(self, context: SqlContext) -> str:
        sql = ["SELECT"]

        projections = [p.to_sql(context) for p in self._projections]
        sql.append(', '.join(projections))

        sql.append("FROM")
        sql.append(f"{self.table_reference.to_sql(context)}")

        if len(self._joins) > 0:
            for join in self._joins:
                sql.append(join.to_sql(context))

        if len(self._predicates) > 0:
            sql.append("WHERE")
            predicates = [p.to_sql(context) for p in self._predicates]
            sql.append(" AND ".join(predicates))

        if len(self._groups) > 0:
            groups = [g.to_sql(context) for g in self._groups]
            sql.append("GROUP BY")
            sql.append(", ".join(groups))

            if self._having_predicates is not None and len(self._having_predicates):
                predicates = [p.to_sql(context) for p in self._having_predicates]
                sql.append("HAVING")
                sql.append(" AND ".join(predicates))

        if self._order_by_items is not None and len(self._order_by_items) > 0:
            sql.append("ORDER BY")
            sql.append(", ".join([o.to_sql(context) for o in self._order_by_items]))

        if self._limit is not None:
            sql.append(self._limit.to_sql(context))

        if self._offset is not None:
            sql.append(self._offset.to_sql(context))

        return " ".join(sql)
