from abc import ABC
from typing import Optional, List, Any

from .sql_generic import SqlNode, SqlNodeInspection, SqlContext, SqlKeyword, SqlOperator, SqlNodeIntrospection, \
    RewritableSqlNode
from ..parser import ast

K_AND = SqlKeyword("AND")
K_OR = SqlKeyword("OR")
K_AS = SqlKeyword("AS")
K_ON = SqlKeyword("ON")

K_LEFT_JOIN = SqlKeyword("LEFT JOIN")
K_RIGHT_JOIN = SqlKeyword("RIGHT JOIN")
K_INNER_JOIN = SqlKeyword("INNER JOIN")
K_FULL_JOIN = SqlKeyword("FULL JOIN")
K_CROSS_JOIN = SqlKeyword("CROSS JOIN")

K_DISTINCT = SqlKeyword("DISTINCT")
K_ORDER_BY = SqlKeyword("ORDER BY")
K_ASC = SqlKeyword("ASC")
K_DESC = SqlKeyword("DESC")
K_LIMIT = SqlKeyword("LIMIT")
K_OFFSET = SqlKeyword("OFFSET")
K_HAVING = SqlKeyword("HAVING")

O_PLUS = SqlOperator("+")
O_MINUS = SqlOperator("-")
O_TIMES = SqlOperator("*")
O_DIVIDE = SqlOperator("/")
O_EQ = SqlOperator("=")
O_GT = SqlOperator(">")
O_LT = SqlOperator("<")
O_GE = SqlOperator(">=")
O_LE = SqlOperator("<=")
O_NE = SqlOperator("!=")

LPAREN = SqlOperator("(")
RPAREN = SqlOperator(")")


class SqlNull(SqlNode):

    def to_sql(self, context: 'SqlContext') -> str:
        return "NULL"


class SqlIdentifier(SqlNode):

    def __init__(self, identifier: str):
        self._identifier = identifier

    @property
    def value(self):
        return self._identifier

    def to_sql(self, context: 'SqlContext') -> str:
        from flatql.parser.ast import Identifier

        if isinstance(self._identifier, Identifier):
            return context.quote_identifier(self._identifier.value())
        return context.quote_identifier(self._identifier)


class SqlQualifiedName(SqlNode):

    def __init__(self, qualified_name: 'ast.QualifiedName'):
        self.qualified_name = qualified_name

    def to_sql(self, context: 'SqlContext') -> str:
        return ".".join([context.quote_identifier(q.value()) for q in self.qualified_name.parts])


class SqlLiteral(SqlNode):

    def __init__(self, literal: 'ast.Literal'):
        self.literal = literal

    def to_sql(self, context: 'SqlContext') -> str:
        from ..parser.ast import LITERAL_TYPE_STRING

        if self.literal.literal_type == LITERAL_TYPE_STRING:
            return context.quote_string_literal(self.literal.value())
        else:
            return self.literal.value()


class SqlArrayLiteral(SqlNode):

    def __init__(self, literal: 'ast.ArrayLiteral'):
        self.literal = literal

    def to_sql(self, context: 'SqlContext') -> str:
        from ..parser.ast import LITERAL_TYPE_STRING, Literal

        def quote(literal: Literal):
            if literal.literal_type == LITERAL_TYPE_STRING:
                return context.quote_string_literal(literal.value())
            else:
                return literal.value()

        return f"[{','.join([quote(l) for l in self.literal.value()])}]"


class SqlColumn(SqlNode, SqlNodeInspection, RewritableSqlNode):

    def __init__(self, table: 'SqlNode', column_name: SqlIdentifier, name: Optional[str] = None,
                 alias: Optional[SqlIdentifier] = None) -> object:
        super(SqlColumn, self).__init__()
        self.table = table
        self.column_name = column_name
        self.name = name
        self._original_alias = alias
        self.alias = alias
        self._rewritten = None

    def inspect(self, introspection: 'SqlNodeIntrospection'):
        introspection.add_column(self)

        if isinstance(self.table, SqlTable):
            introspection.add_table(self.table)

    def rewrite_reference(self, original: SqlNode, rewritten: SqlNode, is_whole_rewrite: Optional[bool] = False):
        if original == self:
            self._rewritten = rewritten
            if self._original_alias is not None:
                self._rewritten.alias = self._original_alias

    def to_sql(self, context: SqlContext) -> str:
        if self._rewritten is not None:
            return self._rewritten.to_sql(context)
        else:
            table_alias = context.alias(self.table).to_sql(context)
            column_alias = "" if not self.alias else f" {K_AS.to_sql(context)} {self.alias.to_sql(context)}"
            sql = [table_alias, ".", self.column_name.to_sql(context)]
            if context.is_show_alias:
                sql.append(column_alias)
            return "".join(sql)

    def copy(self) -> 'SqlNode':
        new = super().copy()
        new._rewritten = None
        return new


class SqlTable(SqlNode):

    def __init__(self, name: str, table_name: SqlNode):
        self.name = name
        self.table_name = table_name

    def to_sql(self, context: SqlContext) -> str:
        table_alias = context.alias(self).to_sql(context)
        table_name = self.table_name.to_sql(context)
        return " ".join([table_name, K_AS.to_sql(context), table_alias])

    def __lt__(self, other):
        return self.name < other.name


class SqlCaseExpression(SqlNode, SqlNodeInspection, RewritableSqlNode):

    def __init__(self, whens_expression: List[SqlNode], else_expression: SqlNode,
                 alias: Optional[SqlIdentifier] = None):
        super(SqlCaseExpression, self).__init__()
        self.whens_expression = whens_expression
        self.else_expression = else_expression
        self.alias = alias
        self._original_alias = alias
        self._rewritten = None

    def inspect(self, introspection: 'SqlNodeIntrospection'):
        _introspection = introspection.create_node_introspection(self)
        for when_expression in self.whens_expression:
            if isinstance(when_expression, SqlNodeInspection):
                when_expression.inspect(_introspection)

        if self.else_expression is not None and isinstance(self.else_expression, SqlNodeInspection):
            self.else_expression.inspect(_introspection)

        introspection.add_case_when_expression(self)

    def rewrite_reference(self, original: SqlNode, rewritten: SqlNode, is_whole_rewrite: Optional[bool] = False):
        if is_whole_rewrite and original == self:
            self._rewritten = rewritten
            if self._original_alias is not None:
                self._rewritten.alias = self._original_alias

    def copy(self) -> 'SqlNode':
        new = super().copy()
        new.whens_expression = self.whens_expression.copy()
        if self.else_expression is not None:
            new.else_expression = self.else_expression.copy()
        new._rewritten = None
        return new

    def to_sql(self, context: 'SqlContext') -> str:
        if self._rewritten is not None:
            return self._rewritten.to_sql(context)
        else:
            sql = ['CASE']
            _context = context.copy(is_show_alias=False)
            for when_expression in self.whens_expression:
                sql.append(when_expression.to_sql(_context))

            if self.else_expression is not None:
                sql.append('ELSE')
                sql.append(self.else_expression.to_sql(_context))
            sql.append("END")

            if self.alias is not None and context.is_show_alias:
                sql.append(K_AS.to_sql(context))
                sql.append(self.alias.to_sql(context))
            return " ".join(sql)


class SqlWhenExpression(SqlNode, SqlNodeInspection, RewritableSqlNode):

    def __init__(self, when_expression: SqlNode, then_expression: SqlNode):
        super(SqlWhenExpression, self).__init__()
        self.when_expression = when_expression
        self.then_expression = then_expression

    def inspect(self, introspection: 'SqlNodeIntrospection'):
        _introspection = introspection.create_node_introspection(self)
        if isinstance(self.when_expression, SqlNodeInspection):
            self.when_expression.inspect(_introspection)

        if isinstance(self.then_expression, SqlNodeInspection):
            self.then_expression.inspect(_introspection)

    def rewrite_reference(self, original: SqlNode, rewritten: SqlNode, is_whole_rewrite: Optional[bool] = False):
        pass

    def copy(self) -> 'SqlNode':
        return super().copy()

    def to_sql(self, context: 'SqlContext') -> str:
        logical_expr = SqlLogicalExpression([self.when_expression])
        return " ".join(['WHEN', logical_expr.to_sql(context), 'THEN', self.then_expression.to_sql(context)])


class SqlPredicate(ABC, SqlNode, SqlNodeInspection):

    def __init__(self, operand: SqlNode, joiner: Optional[SqlKeyword] = K_AND):
        self.operand = operand
        self.joiner = joiner

    def inspect(self, introspection: 'SqlNodeIntrospection'):
        introspection.add_predicate(self)

        if isinstance(self.operand, SqlNodeInspection):
            _introspection = introspection.create_node_introspection(self)
            self.operand.inspect(_introspection)


class SqlComparison(SqlPredicate):

    def __init__(self, operand: SqlNode, operator: SqlOperator, rhs: SqlNode, joiner: Optional[SqlKeyword] = K_AND):
        super(SqlComparison, self).__init__(operand, joiner)
        self.operator = operator
        self.rhs = rhs

    def to_sql(self, context: SqlContext) -> str:
        sql = [self.operand.to_sql(context), self.operator.to_sql(context), self.rhs.to_sql(context)]
        return " ".join(sql)


class SqlIn(SqlPredicate):

    def __init__(self, operand: SqlNode, negated: bool, literals: List[SqlLiteral],
                 joiner: Optional[SqlKeyword] = K_AND):
        super(SqlIn, self).__init__(operand, joiner)
        self.negated = negated
        self.literals = literals

    def to_sql(self, context: SqlContext) -> str:
        literals = [li.to_sql(context) for li in self.literals]
        sql = [self.operand.to_sql(context), "NOT IN" if self.negated else "IN", f'({", ".join(literals)})']
        return " ".join(sql)


class SqlLike(SqlPredicate):

    def __init__(self, operand: SqlNode, negated: bool, string_literal: SqlLiteral,
                 joiner: Optional[SqlKeyword] = K_AND):
        super(SqlLike, self).__init__(operand, joiner)
        self.negated = negated
        self.string_literal = string_literal

    def to_sql(self, context: SqlContext) -> str:
        sql = [self.operand.to_sql(context), "NOT LIKE" if self.negated else "LIKE",
               self.string_literal.to_sql(context)]
        return " ".join(sql)


class SqlIsNull(SqlPredicate):

    def __init__(self, operand: SqlNode, negated: Optional[bool] = False, joiner: Optional[SqlKeyword] = K_AND):
        super(SqlIsNull, self).__init__(operand, joiner)
        self.negated = negated

    def to_sql(self, context: SqlContext) -> str:
        sql = [self.operand.to_sql(context), "IS", "NOT NULL" if self.negated else "NULL"]
        return " ".join(sql)


class SqlBetween(SqlPredicate):

    def __init__(self, operand: SqlNode, negated: bool, literal1: SqlLiteral,
                 literal2: SqlLiteral, joiner: Optional[SqlKeyword] = K_AND):
        super(SqlBetween, self).__init__(operand, joiner)
        self.negated = negated
        self.literal1 = literal1
        self.literal2 = literal2

    def to_sql(self, context: SqlContext) -> str:
        sql = [self.operand.to_sql(context), "NOT BETWEEN" if self.negated else "BETWEEN",
               self.literal1.to_sql(context), "AND", self.literal2.to_sql(context)]
        return " ".join(sql)


class SqlFunctionCall(SqlNode, SqlNodeInspection, RewritableSqlNode):

    def __init__(self, function_name: str, args: List[SqlNode], alias: Optional[SqlNode] = None):
        super(SqlFunctionCall, self).__init__()
        if not isinstance(args, list):
            raise RuntimeError("'args' must be a list")
        self.function_name = function_name
        self.args = args if args is not None else []
        self.alias = alias
        self._rewritten = None
        self._original_alias = alias

    # noinspection PyUnresolvedReferences
    def inspect(self, introspection: 'SqlNodeIntrospection'):
        introspection.add_function_call(self)

        introspectable_args = list(filter(lambda a: isinstance(a, SqlNodeInspection), self.args))
        if len(introspectable_args) > 0:
            _introspection = introspection.create_node_introspection(self)
            for arg in introspectable_args:
                arg.inspect(_introspection)

    # noinspection PyUnresolvedReferences
    def rewrite_reference(self, original: SqlNode, rewritten: SqlNode, is_whole_rewrite: Optional[bool] = False):
        for arg in self.args:
            if arg == original:
                arg = arg.copy()
                arg.rewrite_reference(original, rewritten)
                self.args[self.args.index(arg)] = arg
            elif isinstance(arg, RewritableSqlNode):
                arg.rewrite_reference(original, rewritten)

        if original == self:
            self._rewritten = rewritten
            if self._original_alias is not None:
                self._rewritten.alias = self._original_alias

    def copy(self) -> 'SqlNode':
        new_function_call = super().copy()
        new_function_call.args = [a.copy() for a in self.args]
        new_function_call._rewritten = None
        return new_function_call

    def to_sql(self, context: SqlContext) -> str:
        if self._rewritten is not None:
            return self._rewritten.to_sql(context)
        else:
            _context = context.copy(is_show_alias=False)
            args = [a.to_sql(_context) for a in self.args]
            alias = f" {K_AS.to_sql(context)} {self.alias.to_sql(context)}" if self.alias is not None else ""
            sql = [self.function_name, f'({", ".join(args)})']
            if context.is_show_alias:
                sql.append(alias)
            return "".join(sql)

    def __hash__(self) -> int:
        return super().__hash__()

    def __eq__(self, other: Any) -> bool:
        return self.__class__ == other.__class__ and \
               self.function_name == other.function_name and self.args == other.args


class SqlNormalFunctionCall(SqlFunctionCall):

    def __init__(self, function_name: str, args: List[SqlNode], alias: Optional[SqlNode] = None):
        super(SqlNormalFunctionCall, self).__init__(function_name, args, alias=alias)


class SqlAggFunctionCall(SqlFunctionCall):

    def __init__(self, function_name: str, args: List[SqlNode], alias: Optional[SqlNode] = None):
        super(SqlAggFunctionCall, self).__init__(function_name, args, alias=alias)


class SqlCountDistinctFunctionCall(SqlAggFunctionCall):

    def __init__(self, args: List[SqlNode], alias: Optional[SqlNode] = None):
        super(SqlCountDistinctFunctionCall, self).__init__("COUNT", args, alias=alias)

    def to_sql(self, context: SqlContext) -> str:
        args = [a.to_sql(context) for a in self.args]
        alias = f" {K_AS.to_sql(context)} {self.alias.to_sql(context)}" if self.alias is not None else ""
        sql = [self.function_name, f'({K_DISTINCT} {", ".join(args)})', alias]
        return "".join(sql)


class SqlAdditiveAggrFunctionCall(SqlAggFunctionCall):

    def __init__(self, function_name: str, args: List[SqlNode], alias: Optional[SqlNode] = None):
        super(SqlAdditiveAggrFunctionCall, self).__init__(function_name, args, alias=alias)


class SqlSumFunctionCall(SqlAdditiveAggrFunctionCall):

    def __init__(self, args: List[SqlNode], alias: Optional[SqlNode] = None):
        super(SqlSumFunctionCall, self).__init__("SUM", args, alias=alias)


class SqlCountFunctionCall(SqlAdditiveAggrFunctionCall):

    def __init__(self, args: List[SqlNode], alias: Optional[SqlNode] = None):
        super(SqlCountFunctionCall, self).__init__("COUNT", args, alias=alias)


class Subquery(SqlNode):

    def __init__(self, select: SqlNode):
        self.select = select

    def to_sql(self, context: 'SqlContext') -> str:
        return " ".join([f"({self.select.to_sql(context)})", K_AS.to_sql(context),
                         context.alias(self.select).to_sql(context)])


class SqlLogicalExpression(SqlNode, SqlNodeInspection):

    def __init__(self, predicates: List[SqlPredicate]):
        if not isinstance(predicates, list):
            raise RuntimeError("Predicates must be list of SqlPredicate")
        self.predicates = predicates if predicates is not None else []

    # noinspection PyUnresolvedReferences
    def inspect(self, introspection: 'SqlNodeIntrospection'):
        for predicate in self.predicates:
            predicate.inspect(introspection)

    def append(self, predicate: SqlNode):
        if predicate not in self.predicates:
            self.predicates.append(predicate)
        else:
            pass

    def to_sql(self, context: 'SqlContext') -> str:
        return f" {K_AND.to_sql(context)} ".join([p.to_sql(context) for p in self.predicates])

    def __iter__(self):
        return iter(self.predicates)


class SqlJoin(SqlNode):

    def __init__(self, join_type: SqlKeyword, subclause: SqlNode, on_expr: Optional[SqlNode] = None):
        self.join_type = join_type
        self.subclause = subclause
        self.on_expr = on_expr

    def to_sql(self, context: SqlContext) -> str:
        sql = [self.join_type.to_sql(context), self.subclause.to_sql(context)]
        if self.on_expr is not None:
            sql.append(K_ON.to_sql(context))
            sql.append(self.on_expr.to_sql(context))
        return " ".join(sql)


class SqlOrderByItem(SqlNode, SqlNodeInspection):

    def __init__(self, item: SqlNode, direction: SqlKeyword):
        self.item = item
        self.direction = direction

    def inspect(self, introspection: 'SqlNodeIntrospection'):
        if isinstance(self.item, SqlNodeInspection):
            self.item.inspect(introspection)

    def to_sql(self, context: 'SqlContext') -> str:
        sql = [self.item.to_sql(context), self.direction.to_sql(context)]
        return " ".join(sql)


class SqlLimit(SqlNode):

    def __init__(self, limit: 'ast.Limit'):
        self.limit = SqlLiteral(limit.literal)

    def to_sql(self, context: 'SqlContext') -> str:
        sql = [K_LIMIT.to_sql(context), self.limit.to_sql(context)]
        return " ".join(sql)


class SqlOffset(SqlNode):

    def __init__(self, offset: 'ast.Offset'):
        self.offset = SqlLiteral(offset.literal)

    def to_sql(self, context: 'SqlContext') -> str:
        sql = [K_OFFSET.to_sql(context), self.offset.to_sql(context)]
        return " ".join(sql)


class SqlBinaryExpression(SqlNode, SqlNodeInspection, RewritableSqlNode):

    def __init__(self, lhs: SqlNode, operator: SqlNode, rhs: SqlNode, alias: Optional[SqlIdentifier] = None):
        super(SqlBinaryExpression, self).__init__()
        self.lhs = lhs
        self.operator = operator
        self.rhs = rhs
        self.alias = alias
        self._original_alias = alias
        self._rewritten = None

    def inspect(self, introspection: 'SqlNodeIntrospection'):
        _introspection = None
        if isinstance(self.lhs, SqlNodeInspection):
            _introspection = introspection.create_node_introspection(self)
            self.lhs.inspect(_introspection)

        if isinstance(self.rhs, SqlNodeInspection):
            if _introspection is None:
                _introspection = introspection.create_node_introspection(self)
            self.rhs.inspect(_introspection)

        introspection.add_binary_expression(self)

    # noinspection PyUnresolvedReferences
    def rewrite_reference(self, original: SqlNode, rewritten: SqlNode, is_whole_rewrite: Optional[bool] = False):
        if isinstance(self.lhs, RewritableSqlNode):
            if self.lhs == original:
                new_lhs = self.lhs.copy()
                new_lhs.rewrite_reference(original, rewritten)
                self.lhs = new_lhs
            else:
                self.lhs.rewrite_reference(original, rewritten)

        if isinstance(self.rhs, RewritableSqlNode):
            if self.rhs == original:
                new_rhs = self.rhs.copy()
                new_rhs.rewrite_reference(original, rewritten)
                self.rhs = new_rhs
            else:
                self.rhs.rewrite_reference(original, rewritten)

        if is_whole_rewrite and original == self:
            self._rewritten = rewritten
            if self._original_alias is not None:
                self._rewritten.alias = self._original_alias

    def copy(self) -> 'SqlNode':
        new = super().copy()
        new._rewritten = None
        new.lhs = self.lhs.copy()
        new.rhs = self.rhs.copy()
        return new

    def to_sql(self, context: 'SqlContext') -> str:
        if self._rewritten is not None:
            return self._rewritten.to_sql(context)
        else:
            context = context.copy(is_show_alias=False)
            sql = [self.lhs.to_sql(context), self.operator.to_sql(context), self.rhs.to_sql(context)]
            if self.alias is not None:
                sql.append(K_AS.to_sql(context))
                sql.append(self.alias.to_sql(context))
            return " ".join(sql)


class SqlParenExpression(SqlNode, SqlNodeInspection, RewritableSqlNode):

    def __init__(self, expression: SqlNode, alias: Optional[SqlIdentifier] = None):
        super(SqlParenExpression, self).__init__()
        self.expression = expression
        self.alias = alias
        self._original_alias = alias

    def inspect(self, introspection: 'SqlNodeIntrospection'):
        introspection.add_binary_expression(self)

        if isinstance(self.expression, SqlNodeInspection):
            _introspection = introspection.create_node_introspection(self)
            self.expression.inspect(_introspection)

    # noinspection PyUnresolvedReferences
    def rewrite_reference(self, original: SqlNode, rewritten: SqlNode, is_whole_rewrite: Optional[bool] = False):
        if isinstance(self.expression, RewritableSqlNode):
            if self.expression == original:
                expression = self.expression.copy()
                expression.rewrite_reference(original, rewritten)
                self.expression = expression
            else:
                self.expression.rewrite_reference(original, rewritten)

        if original == self:
            self._rewritten = rewritten
            if self._original_alias is not None:
                self._rewritten.alias = self._original_alias

    def to_sql(self, context: 'SqlContext') -> str:
        context = context.copy(is_show_alias=False)
        sql = [f"({self.expression.to_sql(context)})"]
        if self.alias is not None:
            sql.append(K_AS.to_sql(context))
            sql.append(self.alias.to_sql(context))
        return " ".join(sql)

    def copy(self) -> 'SqlNode':
        new = super().copy()
        new._rewritten = None
        new.expression = self.expression.copy()
        return new


def double_quoted_identifier(identifier: str) -> str:
    return f"\"{identifier}\""


def backquoted_identifier(identifier: str) -> str:
    return f"`{identifier}`"


def quote_string_literal(string_literal: str) -> str:
    return f"'{string_literal}'"
