from abc import ABCMeta, abstractmethod, ABC
from collections import defaultdict
from copy import copy
from typing import Set, List, Dict, Optional, Any

from .frozen import immutable_property
from ..schemas import View, Table


class SqlNodeInspection(metaclass=ABCMeta):

    @abstractmethod
    def inspect(self, introspection: 'SqlNodeIntrospection'): pass


class SqlContext(metaclass=ABCMeta):

    def __init__(self,
                 view: Optional['View'] = None,
                 tables_dict: Optional[Dict[str, 'SqlNode']] = None,
                 join_conditions: Optional[Dict[str, Dict[str, 'SqlNode']]] = None,
                 table_alias_prefix: Optional[str] = "ta_",
                 column_alias_prefix: Optional[str] = "co_",
                 subquery_alias_prefix: Optional[str] = "qu_"):
        self._view = view
        self._tables_dict = tables_dict
        self._join_conditions = join_conditions
        self._table_alias_prefix = table_alias_prefix
        self._column_alias_prefix = column_alias_prefix
        self._subquery_alias_prefix = subquery_alias_prefix
        self._table_list, self._column_list, self._subquery_list = [], [], []
        self._is_show_alias = True

    @property
    def is_show_alias(self):
        return self._is_show_alias

    def alias(self, node: 'SqlNode') -> 'SqlNode':
        from .select import Select
        from .sql import SqlTable, SqlIdentifier

        if isinstance(node, SqlTable):
            if node not in self._table_list:
                self._table_list.append(node)
            return SqlIdentifier(f"{self._table_alias_prefix}{self._table_list.index(node)}")
        elif isinstance(node, Select):
            if node not in self._subquery_list:
                self._subquery_list.append(node)
            return SqlIdentifier(f"{self._subquery_alias_prefix}{self._subquery_list.index(node)}")
        else:
            if node not in self._column_list:
                self._column_list.append(node)
            return SqlIdentifier(f"{self._column_alias_prefix}{self._column_list.index(node)}")

    # noinspection PyUnresolvedReferences
    def create_table_primary_key(self, table: 'SqlNode'):
        from .sql import SqlColumn, SqlIdentifier

        primary_key = self.get_table_schema(table.name).primary_key
        return SqlColumn(table, SqlIdentifier(primary_key))

    def get_main_table(self, tables_name: List[str]):
        from flatql.rewriter.algorithm import decide_main_table
        return decide_main_table(
            tables_name=tables_name,
            relation_graph=self.get_relation_graph()
        )

    def get_relation_graph(self):
        return self._view.get_relation_graph()

    # noinspection PyUnresolvedReferences
    def detect_involved_tables(self, tables_name: List[str], tables: Set['SqlNode']) -> Dict[str, 'SqlNode']:
        from flatql.rewriter.algorithm import detect_involved_tables

        queried_tables_name = list(tables_name)
        relation_graph = self.get_relation_graph()
        _, involved_tables_name = detect_involved_tables(
            queried_tables_name[0], queried_tables_name[1:], {queried_tables_name[0]}, relation_graph)

        for table_name in involved_tables_name:
            if not (table_name in tables):
                tables.add(self.get_sql_table(table_name))

        return {t.name: t for t in sorted(list(tables))}

    def get_join_condition(self, parent_table_name: str, table_name: str) -> 'SqlNode':
        join_condition = self._join_conditions.get(parent_table_name).get(table_name)
        if join_condition is None:
            join_condition = self._join_conditions.get(table_name).get(parent_table_name)
        return join_condition

    def get_table_schema(self, name: str) -> 'Table':
        return self._view.get_table(name)

    def get_sql_table(self, name: str) -> 'SqlNode':
        return self._tables_dict.get(name)

    def quote_string_literal(self, string_literal: str) -> str:
        return quote_string_literal(string_literal)

    def quote_identifier(self, identifier: str) -> str:
        return double_quoted_identifier(identifier)

    def copy(self, is_show_alias=True):
        new = copy(self)
        new._is_show_alias = is_show_alias
        return new


class SqlContextWrapper(SqlContext):

    def __init__(self, context: SqlContext):
        super().__init__(
            view=context._view,
            tables_dict=context._tables_dict,
            join_conditions=context._join_conditions
        )


class SqlNode(metaclass=ABCMeta):

    @abstractmethod
    def to_sql(self, context: 'SqlContext') -> str:
        pass

    def copy(self) -> 'SqlNode':
        return copy(self)

    def __lt__(self, other):
        return self.__hash__() < other.__hash__()

    def __hash__(self) -> int:
        return super().__hash__()

    def __eq__(self, other: Any) -> bool:
        return other.__class__ == self.__class__ and self.__dict__ == other.__dict__


class RewritableSqlNode(ABC):

    def __init__(self):
        self._original_alias = None

    @abstractmethod
    def rewrite_reference(self, original: SqlNode, rewritten: SqlNode, is_whole_rewrite: Optional[bool] = False): pass

    @property
    def original_alias(self):
        return self._original_alias


class SqlNodeIntrospection:

    def __init__(self, context: SqlContext, projections: List[SqlNode], predicates: List[SqlNode]):
        self._context = context

        self._binary_expressions = list()
        self._case_when_expression = list()
        self._columns: List[SqlNode] = list()
        self._function_calls: List[SqlNode] = list()
        self._tables: Set[SqlNode] = set()
        self._projections: List[SqlNode] = projections if projections is not None else []
        self._predicates: List[SqlNode] = predicates if predicates is not None else []
        self._node_introspections: Dict[SqlNode, SqlNodeIntrospection] = dict()

        self._inspect(self._projections)
        self._inspect(self._predicates)

        if len(self.tables) > 0:
            self._detect_involved_tables()

    @immutable_property
    def dimensions(self) -> List[SqlNode]:
        from flatql.rewriter.sql import SqlColumn, SqlBinaryExpression, SqlCaseExpression, \
            SqlFunctionCall, SqlAggFunctionCall

        def filter_dimensions(p: SqlNode):
            if isinstance(p, SqlColumn):
                return True
            elif not isinstance(p, SqlAggFunctionCall) and isinstance(p, SqlFunctionCall):
                return not self.get_node_introspection(p).has_aggr_function_calls()
            elif isinstance(p, SqlBinaryExpression) or isinstance(p, SqlCaseExpression):
                return not self.get_node_introspection(p).has_aggr_function_calls()
            return False

        return list(filter(lambda p: filter_dimensions(p), self.projections))

    @immutable_property
    def projections(self) -> List[SqlNode]:
        return self._projections

    @property
    def predicates(self) -> SqlNode:
        return self._predicates

    @immutable_property
    def where_predicates(self) -> Dict[SqlNode, SqlNode]: # SqlTable, SqlNode
        from .sql import SqlPredicate, SqlFunctionCall, SqlAggFunctionCall

        where_predicates = defaultdict(list)
        for predicate in self._predicates:
            if isinstance(predicate, SqlFunctionCall) or \
                    (isinstance(predicate, SqlPredicate) and not isinstance(predicate.operand, SqlAggFunctionCall)):
                _introspection = SqlNodeIntrospection.inspect(self._context, [[predicate]])
                where_predicates[_introspection.table].append(predicate)
            # TODO Ignore having predicate now.
        return where_predicates

    @immutable_property
    def tables(self) -> List[SqlNode]:
        tables = set(self._tables)
        for introspection in self._node_introspections.values():
            tables = tables.union(introspection.tables)
        return sorted(list(tables))

    @immutable_property
    def tables_dict(self) -> Dict[str, SqlNode]:
        return {t.name: t for t in self.tables}

    @immutable_property
    def tables_name(self) -> List[str]:
        return sorted([t.name for t in self.tables])

    @property
    def table(self) -> SqlNode:
        if len(self.tables) > 1:
            raise RuntimeError(f"Too many tables in single table query: {self.tables}")
        return self.tables[0]

    @immutable_property
    def function_calls(self) -> List[SqlNode]:
        function_calls = self._function_calls
        for introspection in self._node_introspections.values():
            function_calls = function_calls + introspection.function_calls
        return function_calls

    @immutable_property
    def aggr_function_calls(self) -> Dict[SqlNode, SqlNode]:  # Dict[SqlTable, SqlNode]
        from flatql.rewriter.sql import SqlAggFunctionCall

        aggr_function_calls = defaultdict(list)
        for function_call in self.function_calls:
            if isinstance(function_call, SqlAggFunctionCall):
                introspection = self.get_node_introspection(function_call)
                tables = introspection.tables
                if len(tables) > 1:
                    raise RuntimeError(f"Aggregate function calls only can have one table.")
                else:
                    if tables[0] not in aggr_function_calls or \
                            function_call not in aggr_function_calls[tables[0]]:
                        aggr_function_calls[tables[0]].append(function_call)

        return aggr_function_calls

    def is_whole_rewrite(self, node: SqlNode) -> bool:
        from .sql import SqlAggFunctionCall
        introspection = self.get_node_introspection(node)
        if introspection is None:
            return False
        return (not isinstance(node, SqlAggFunctionCall)) and \
               (not self.get_node_introspection(node).has_aggr_function_calls())

    def get_predicates(self, tables: List[SqlNode]) -> List[SqlNode]:
        return list(filter(lambda p: self.get_node_introspection(p).table in tables, self.predicates))

    def get_node_introspection(self, node: SqlNode) -> 'SqlNodeIntrospection':
        introspection = self._node_introspections.get(node)
        if introspection is not None:
            return introspection
        else:
            for _introspection in self._node_introspections.values():
                introspection = _introspection.get_node_introspection(node)
                if introspection is not None:
                    return introspection

    def has_dimensions(self) -> bool:
        return len(self.dimensions) > 0

    def has_multiple_tables(self) -> bool:
        return len(self.tables) > 1

    def has_aggr_function_calls(self, function_calls: Optional[List[SqlNode]] = None) -> bool:
        from .sql import SqlAggFunctionCall

        function_calls = function_calls if function_calls is not None else self.function_calls
        aggr_function_calls = list(filter(lambda f: isinstance(f, SqlAggFunctionCall), function_calls))
        return len(aggr_function_calls) > 0

    def has_additive_aggr_function_calls(self, function_calls: Optional[List[SqlNode]] = None) -> bool:
        from .sql import SqlAdditiveAggrFunctionCall
        function_calls = function_calls if function_calls is not None else self.function_calls
        additive_aggr_function_calls = list(
            filter(lambda f: isinstance(f, SqlAdditiveAggrFunctionCall), function_calls))
        return len(additive_aggr_function_calls) > 0

    def add_table(self, table: SqlNode):
        self._tables.add(table)

    def add_binary_expression(self, expression: SqlNode):
        if expression not in self._binary_expressions:
            self._binary_expressions.append(expression)

    def add_case_when_expression(self, expression: SqlNode):
        if expression not in self._case_when_expression:
            self._case_when_expression.append(expression)

    def add_column(self, node: SqlNode):
        if node not in self._columns:
            self._columns.append(node)

    def add_predicate(self, predicate: SqlNode):
        if predicate not in self._predicates:
            self._predicates.append(predicate)

    def add_function_call(self, function_call: SqlNode):
        if function_call not in self._function_calls:
            self._function_calls.append(function_call)

    def create_node_introspection(self, node: SqlNode) -> 'SqlNodeIntrospection':
        introspection = SqlNodeIntrospection(self._context, [], [])
        self._node_introspections[node] = introspection
        return introspection

    def _inspect(self, nodes: List[SqlNode]):
        for node in nodes:
            if isinstance(node, SqlNodeInspection):
                node.inspect(self)

    def _detect_involved_tables(self):
        from flatql.rewriter.algorithm import detect_involved_tables

        queried_tables_name = self.tables_name
        relation_graph = self._context.get_relation_graph()
        _, involved_tables_name = detect_involved_tables(
            queried_tables_name[0], queried_tables_name[1:], {queried_tables_name[0]}, relation_graph)

        for table_name in involved_tables_name:
            if not (table_name in self.tables):
                self._tables.add(self._context.get_sql_table(table_name))

    @staticmethod
    def inspect(context: SqlContext, nodes_list: List[List[SqlNode]]) -> 'SqlNodeIntrospection':
        introspection = SqlNodeIntrospection(context, [], [])
        for nodes in nodes_list:
            if nodes is not None:
                introspection._inspect(nodes)

        if len(introspection.tables) > 0:
            introspection._detect_involved_tables()
        return introspection


class SqlKeyword(SqlNode):

    def __init__(self, keyword: str):
        self.keyword = keyword

    def to_sql(self, context: 'SqlContext') -> str:
        return self.keyword


class SqlOperator(SqlNode):

    def __init__(self, text: str):
        self.text = text

    def to_sql(self, context: 'SqlContext') -> str:
        return self.text


def double_quoted_identifier(identifier: str) -> str:
    return f"\"{identifier}\""


def backquoted_identifier(identifier: str) -> str:
    return f"`{identifier}`"


def quote_string_literal(string_literal: str) -> str:
    return f"'{string_literal}'"
