from typing import List, Dict, Optional, Callable

from . import query
from .algorithm import detect_involved_tables
from .function_call_rewriter import FunctionCallRewriterFactory, rewrite_sql_function_call
from .sql import SqlIdentifier, SqlLiteral, SqlTable, SqlColumn, \
    SqlKeyword, SqlComparison, SqlOperator, SqlIn, SqlBetween, SqlLike, SqlIsNull, \
    SqlLimit, SqlOffset, SqlOrderByItem, SqlBinaryExpression, SqlParenExpression, SqlNode, \
    SqlLogicalExpression, SqlQualifiedName, SqlContext, SqlNull, SqlCaseExpression, \
    SqlWhenExpression, SqlArrayLiteral
from .sql_generic import SqlNodeIntrospection
from .. import schemas
from ..parser import ast


class SqlRewriter:

    def __init__(self, schema: schemas.Schema,
                 sql_predicate_appender: Optional[query.SqlPredicateAppender] = None,
                 function_call_rewriter_factory: Optional[FunctionCallRewriterFactory] = None):
        self._view = None
        self._schema = schema
        self._tables_dict = {}
        self._projections = None
        self._predicates = None
        self._order_by_items = None
        self._limit = None
        self._offset = None
        self._sql_predicate_appender = sql_predicate_appender
        self._function_call_rewriter_factory = function_call_rewriter_factory

    def set_view(self, view: 'ast.View'):
        self._view = self._schema.get_view(view.name.value())

    def set_projections(self, projections: List['SqlNode']):
        self._projections = projections

    def set_predicates(self, predicates: SqlLogicalExpression):
        self._predicates = predicates

    def set_order_by_items(self, order_by_items: List[SqlOrderByItem]):
        self._order_by_items = order_by_items

    def set_limit(self, limit: SqlNode):
        self._limit = limit

    def set_offset(self, offset: SqlNode):
        self._offset = offset

    def rewrite_identifier(self, identifier: 'ast.Identifier') -> 'SqlNode':
        return SqlIdentifier(identifier.value())

    def rewrite_literal(self, literal: 'ast.Literal') -> 'SqlNode':
        return SqlLiteral(literal)

    def rewrite_array_literal(self, literal: 'ast.ArrayLiteral') -> 'SqlNode':
        return SqlArrayLiteral(literal)

    # noinspection PyUnresolvedReferences
    def rewrite_column(self, column: 'ast.Column') -> 'SqlNode':
        sql_table = column.table.rewrite(self)
        raw_column_name = column.name.rewrite(self)
        column_alias = column.alias.rewrite(self) if column.alias is not None else None

        table_schema = self._view.get_table(sql_table.name)
        db_column_name = table_schema.get_column(raw_column_name.value).db_column_name

        return SqlColumn(sql_table, SqlIdentifier(db_column_name), name=column.name.value(), alias=column_alias)

    def rewrite_function_call(self, function_call: 'ast.FunctionCall') -> 'SqlNode':
        function_name = function_call.name
        args = [a.rewrite(self) for a in function_call.args]
        alias = function_call.alias.rewrite(self) if function_call.alias is not None else None
        function_call_rewriter = self._get_function_call_rewriter(function_name)

        if function_call_rewriter is not None:
            return function_call_rewriter(args, alias=alias)
        else:
            return rewrite_sql_function_call(function_call.name, args, alias)

    def rewrite_comparison(self, comparison: 'ast.Comparison') -> 'SqlNode':
        sql_operand = comparison.operand.rewrite(self)
        sql_operator = SqlOperator(comparison.operator)
        sql_rhs = comparison.literal.rewrite(self)
        return SqlComparison(sql_operand, sql_operator, sql_rhs)

    def rewrite_in(self, in_: 'ast.In') -> 'SqlNode':
        sql_operand = in_.operand.rewrite(self)
        literals = [a.rewrite(self) for a in in_.literals]
        return SqlIn(sql_operand, in_.negated, literals)

    def rewrite_between(self, between: 'ast.Between') -> 'SqlNode':
        sql_operand = between.operand.rewrite(self)
        literal1 = between.literal1.rewrite(self)
        literal2 = between.literal2.rewrite(self)
        return SqlBetween(sql_operand, between.negated, literal1, literal2)

    def rewrite_like(self, like: 'ast.Like') -> 'SqlNode':
        sql_operand = like.operand.rewrite(self)
        literal = like.string_literal.rewrite(self)
        return SqlLike(sql_operand, like.negated, literal)

    def rewrite_is_null(self, is_null: 'ast.IsNull') -> 'SqlNode':
        sql_operand = is_null.operand.rewrite(self)
        return SqlIsNull(sql_operand, is_null.negated)

    def rewrite_order_by_item(self, order_by_item: 'ast.OrderByItem') -> 'SqlNode':
        item = order_by_item.node.rewrite(self)
        direction = SqlKeyword(order_by_item.direction)
        return SqlOrderByItem(item, direction)

    def rewrite_limit(self, limit: 'ast.Limit') -> 'SqlNode':
        return SqlLimit(limit)

    def rewrite_offset(self, offset: 'ast.Offset') -> 'SqlNode':
        return SqlOffset(offset)

    def rewrite_table(self, table: 'ast.Table') -> 'SqlNode':
        from flatql.parser.flatql_parser import parse_qualified_name
        if table.name.value() not in self._tables_dict:
            table_schema = self._view.get_table(table.name.value())
            self._tables_dict[table.name.value()] = SqlTable(table.name.value(), SqlQualifiedName(
                parse_qualified_name(table_schema.table_name)))

        return self._tables_dict[table.name.value()]

    def rewrite_binary_expression(self, binary_expression: 'ast.BinaryExpression') -> 'SqlNode':
        lhs = binary_expression.lhs.rewrite(self)
        operator = SqlOperator(binary_expression.operator)
        rhs = binary_expression.rhs.rewrite(self)
        alias = binary_expression.alias.rewrite(self) if binary_expression.alias is not None else None
        return SqlBinaryExpression(lhs, operator, rhs, alias)

    def rewrite_paren_expression(self, paren_expression: 'ast.ParenExpression') -> 'SqlNode':
        alias = paren_expression.alias.rewrite(self) if paren_expression.alias is not None else None
        paren_expression = SqlParenExpression(paren_expression.binary_expr.rewrite(self), alias)
        return paren_expression

    def rewrite_null(self, _: 'ast.Null') -> 'SqlNode':
        return SqlNull()

    def rewrite_case_expression(self, case: 'ast.CaseExpression') -> 'SqlNode':
        whens_expression = [w.rewrite(self) for w in case.whens_expressions]
        else_expression = case.else_expression.rewrite(self) if case.else_expression is not None else None
        alias = case.alias.rewrite(self) if case.alias is not None else None
        return SqlCaseExpression(whens_expression, else_expression, alias)

    def rewrite_logical_expression(self, logical_expr: 'ast.LogicalExpression') -> SqlNode:
        return SqlLogicalExpression([b.rewrite(self) for b in logical_expr.boolean_expressions])

    def rewrite_when_expression(self, when: 'ast.WhenExpression') -> 'SqlNode':
        boolean_expressions = when.logical_expression.rewrite(self)
        then_expression = when.then_expression.rewrite(self)
        return SqlWhenExpression(boolean_expressions, then_expression)

    def to_sql(self, context: Optional[SqlContext] = None) -> str:
        if context is None:
            context = self._create_sql_context()
        introspection = SqlNodeIntrospection(context, self._projections, self._predicates)

        return query.create_query_tree(
            context=context,
            introspection=introspection,
            order_by_items=self._order_by_items,
            limit=self._limit,
            offset=self._offset,
            sql_predicate_appender=self._sql_predicate_appender
        ).to_sql(context)

    def rewrite_join_conditions(self) -> Dict[str, Dict[str, SqlNode]]:
        from flatql.parser.flatql_parser import parse_join_condition

        queried_tables_name = list(self._view.table_dict.keys())
        _, involved_tables_name = detect_involved_tables(
            queried_tables_name[0], queried_tables_name[1:], {queried_tables_name[0]}, self._view.get_relation_graph())

        join_conditions = dict()
        for table_name in involved_tables_name:
            join_conditions[table_name] = dict()
            table: schemas.Table = self._view.get_table(table_name)
            for join in table.joins:
                if join.target in involved_tables_name:
                    predicates = parse_join_condition(join.on_expr)
                    join_condition = SqlLogicalExpression([p.rewrite(self) for p in predicates])
                    join_conditions[table.name][join.target] = join_condition

        return join_conditions

    def _get_function_call_rewriter(self, function_name: str) -> Callable:
        if self._function_call_rewriter_factory is not None:
            return self._function_call_rewriter_factory.get_rewriter(function_name)
        return None

    def _create_sql_context(self):
        return SqlContext(
            view=self._view,
            tables_dict=self._tables_dict,
            join_conditions=self.rewrite_join_conditions()
        )
