from enum import Enum
from typing import List, Optional, Dict, Any

import schema
from schema import Or


class SchemaException(Exception):
    pass


class SchemaParseException(SchemaException):
    """Error should be raised when invalid schema definition occurs."""
    pass


class EnumBase(Enum):

    @classmethod
    def value_of(cls, name: str) -> Enum:
        if not cls.has(name):
            raise SchemaParseException(f"Enum '{name}' is not in {cls}")
        return cls.__dict__['_member_map_'][name]

    @classmethod
    def has(cls, name: str) -> bool:
        return name in cls.__dict__['_member_names_']

    @classmethod
    def candidates(cls) -> List[str]:
        return cls.__dict__['_member_names_']


class ColumnType(EnumBase):
    ATTRIBUTE = "ATTRIBUTE"
    MEASURE = "MEASURE"


class DataType(EnumBase):
    BOOL = "BOOL"
    DATE = "DATE"
    DATETIME = "DATETIME"
    INTEGER = "INTEGER"
    REAL = "REAL"
    STRING = "STRING"
    ARRAY = "ARRAY"
    MAP = "MAP"


class AggregateType(EnumBase):
    SUM = "SUM"
    AVG = "AVG"
    MIN = "MIN"
    MAX = "MAX"
    COUNT = "COUNT"
    COUNT_DISTINCT = "COUNT_DISTINCT"
    STD_DEVIATION = "STD_DEVIATION"
    VARIANCE = "VARIANCE"


class JoinType(EnumBase):
    LEFT_JOIN = "LEFT OUTER JOIN"
    RIGHT_JOIN = "RIGHT OUTER JOIN"
    INNER_JOIN = "INNER JOIN"
    FULL_JOIN = "FULL OUTER JOIN"


class Column:

    def __init__(self,
                 name: str,
                 db_column_name: str,
                 data_type: DataType,
                 column_type: Optional[ColumnType] = ColumnType.ATTRIBUTE,
                 is_hidden: Optional[bool] = False,
                 aggregations: Optional[List[AggregateType]] = None):
        self.name = name
        self.db_column_name = db_column_name
        self.data_type = data_type
        self.column_type = column_type
        self.is_hidden = is_hidden
        self.aggregations = aggregations if aggregations is not None else []


class Join:

    def __init__(self, target: str, on_expr: str, join_type: Optional[str] = None,
                 is_one_to_one: Optional[bool] = False):
        self.target = target
        self.on_expr = on_expr
        self.join_type = JoinType.value_of(join_type).value if join_type is not None else "INNER JOIN"
        self.is_one_to_one = is_one_to_one


class Table:

    def __init__(self, name: str, table_name: str, primary_key: str, columns: List[Column], joins: List[Join]):
        self.name = name
        self.table_name = table_name
        self.primary_key = primary_key
        self.columns = columns
        self.column_dict = {}
        self.column_dict2 = {}
        self.join_dict = {}
        self.joins = joins

        for column in columns:
            if column.name in self.column_dict:
                raise SchemaParseException(f"Duplicated column name: '{column.name}'")
            self.column_dict[column.name] = column
            self.column_dict2[column.db_column_name] = column

        for join in joins:
            self.join_dict[join.target] = join

    def get_column(self, name: str) -> Column:
        if name not in self.column_dict:
            raise SchemaException(f"Cannot find column '{name}' in table '{self.name}[{self.table_name}]'")
        return self.column_dict[name]

    def get_column_by_db_column_name(self, db_column_name: str) -> Column:
        if db_column_name not in self.column_dict2:
            raise SchemaException(f"Cannot find column '{db_column_name}' in table '{self.name}[{self.table_name}]'")
        return self.column_dict2[db_column_name]

    def has_join(self, target_name: str) -> bool:
        return target_name in self.join_dict

    def get_join(self, target_name: str) -> Join:
        if target_name not in self.join_dict:
            raise SchemaException(f"Cannot find join target '{target_name}' in table '{self.name}'")
        return self.join_dict[target_name]


class View:

    def __init__(self, name: str, tables: List[Table]):
        self.name = name
        self.tables = tables
        self.table_dict = {}
        self._children_map = None
        self._children_map_weighted = None

        for table in self.tables:
            if table.name in self.table_dict:
                raise SchemaParseException(f"Duplicated table name: '{table.name}'")
            self.table_dict[table.name] = table

    def get_relation_graph_weighted(self) -> Dict:
        if self._children_map_weighted is None:
            self._children_map_weighted = {table.name: set() for table in self.tables}
            for table in self.tables:
                for join in table.joins:
                    self._children_map_weighted[table.name].add((join.target, 1))
                    self._children_map_weighted[join.target].add((table.name, 1))
        return self._children_map_weighted

    def get_relation_graph(self) -> Dict:
        if self._children_map is None:
            self._children_map = {table.name: set() for table in self.tables}
            for table in self.tables:
                for join in table.joins:
                    self._children_map[table.name].add(join.target)
                    self._children_map[join.target].add(table.name)

        return self._children_map

    def get_table(self, name: str) -> Table:
        if name not in self.table_dict:
            raise SchemaException(f"Cannot find table '{name}' in view '{self.name}'")
        return self.table_dict.get(name)


class Schema:

    def __init__(self, views: List):
        self.views = views
        self._view_dict = {}

        for view in self.views:
            self._view_dict[view.name] = view

    def get_view(self, view_name: str) -> View:
        if view_name not in self._view_dict:
            raise SchemaException(f"Cannot find view '{view_name}'")
        return self._view_dict.get(view_name)

    def get_table(self, view_name: str, table_name: str) -> Table:
        if view_name not in self._view_dict:
            raise SchemaException(f"Cannot find view '{view_name}'")
        return self._view_dict.get(view_name).get_table(table_name)


schema_definition = schema.Schema({
    'views': [
        {
            'name': str,
            "tables": [
                {
                    'name': str,
                    'table_name': str,
                    'primary_key': str,
                    "columns": [
                        {
                            "name": str,
                            "db_column_name": str,
                            "column_type": Or(*ColumnType.candidates(),
                                              error="Invalid column type: '{0}', candidates: [%s]" % (
                                                  ', '.join(ColumnType.candidates()))),
                            "data_type": Or(*DataType.candidates(),
                                            error="Invalid data type: '{0}', candidates: [%s]" % (
                                                ', '.join(DataType.candidates()))),
                            schema.Optional("is_hidden"): bool,
                            schema.Optional("aggregations"): [Or(*AggregateType.candidates(),
                                                                 error="Invalid aggregate type: '{0}', candidates: [%s]" % (
                                                                     ', '.join(AggregateType.candidates())))]
                        }
                    ],
                    schema.Optional("joins"): [
                        {
                            schema.Optional("type"): Or(*JoinType.candidates(),
                                                        error="Invalid join type: '{0}', candidates: [%s]" % (
                                                            ', '.join(JoinType.candidates()))),
                            "target": str,
                            "on_expr": str,
                            schema.Optional("is_one_to_one"): bool
                        }
                    ]
                }
            ]
        }
    ]
})


def _get_files_by_suffix(path: str, suffix: str):
    import os
    has_suffix_files = set()
    for root, dirs, files in os.walk(path):
        for file in files:
            file_path = os.path.join(root, file)
            if file_path.endswith(suffix):
                has_suffix_files.add(file_path)
    return has_suffix_files


def _parse_raw_attr(raw: Dict, name: str, default_value: Any):
    return raw[name] if name in raw else default_value


def _parse_column(raw_column: Dict) -> Column:
    return Column(raw_column["name"],
                  raw_column["db_column_name"],
                  raw_column["data_type"],
                  raw_column["column_type"],
                  is_hidden=_parse_raw_attr(raw_column, 'is_hidden', False),
                  aggregations=_parse_raw_attr(raw_column, 'aggregations', []),
                  )


def _parse_joins(raw_join: List[Dict]) -> List[Join]:
    joins = []
    for r in raw_join:
        joins.append(
            Join(
                r["target"],
                r["on_expr"],
                join_type=_parse_raw_attr(r, 'type', "INNER_JOIN"),
                is_one_to_one=_parse_raw_attr(r, 'is_one_to_one', False)
            )
        )
    return joins


def _parse_table(raw_table: Dict) -> Table:
    columns: List[Column] = [_parse_column(raw_column) for raw_column in raw_table["columns"]]
    joins: List[Join] = _parse_joins(raw_table.get("joins")) if "joins" in raw_table else []
    return Table(raw_table["name"], raw_table["table_name"], raw_table["primary_key"], columns, joins)


def _parse_view(raw_view: Dict) -> View:
    return View(raw_view["name"], [_parse_table(raw_column) for raw_column in raw_view["tables"]])


def _parse_views(raw_schema: Dict) -> Schema:
    return [_parse_view(v) for v in raw_schema["views"]]
